#version 450
/*
 * Copyright 2023 TU Wien, Institute of Visual Computing & Human-Centered Technology.
 * This file is part of the GCG Lab Framework and must not be redistributed.
 *
 * Original version created by Lukas Gersthofer and Bernhard Steiner.
 * Vulkan edition created by Johannes Unterguggenberger (junt@cg.tuwien.ac.at).
 */
 
layout (binding = 0) uniform UniformBuffer {
	vec4 color;
	mat4 modelMatrix;
	mat4 modelMatrixForNormals;
	mat4 viewProjMatrix;
	vec4 cameraPosition;
	vec4 illumination; // ka, kd, ks, alpha
	ivec4 userInput;

	float time;
    float impactX;
    float impactY;
    float pad3;
} ub_data;

layout (binding = 1) uniform DirectionalLight {
	vec4 color;
	vec4 direction;
} dl_data;

layout (binding = 2) uniform PointLight {
	vec4 color;
	vec4 position;
	vec4 attenuation;
} pl_data;

layout (binding = 3) uniform sampler2D diffuse_texture;

layout (binding = 4) uniform sampler2D normal_map;

layout (location = 0) in VertexData {
	vec3 positionWorld;
	vec3 normalWorld;
	vec2 textureCoordinates;
} frag_in;

layout (location = 0) out vec4 out_color;

vec3 phong(vec3 n, vec3 l, vec3 v, vec3 diffuseC, float diffuseF, vec3 specularC, float specularF, float alpha, bool attenuate, vec3 attenuation) {
	float d = length(l);
	l = normalize(l);
	float att = 1.0;	
	if(attenuate) {
		att = 1.0f / (attenuation.x + d * attenuation.y + d * d * attenuation.z);
	}
	vec3 r = reflect(-l, n);
	return (diffuseF * diffuseC * max(0, dot(n, l)) + specularF * specularC * pow(max(0, dot(r, v)), alpha)) * att; 
}

void main() {	
	vec3 n = normalize(frag_in.normalWorld);
	vec3 v = normalize(frag_in.positionWorld - ub_data.cameraPosition.xyz);

	vec3 diffuseColor;
	float alpha = 1.0;

	if (ub_data.userInput[2] == 1) {
		// Water pass
		diffuseColor = vec3(0.05, 0.25, 0.4); // watery color
		float lakeRadius = 2.0;

		float distFromCenter = length(
            vec2(
                frag_in.positionWorld.x - ub_data.impactX,
                (frag_in.positionWorld.z - 3.0)  - ub_data.impactY
            )
        );

		// We make alpha = 1.0 at the center, and fade to 0.0 at lakeRadius and beyond.
        float t = distFromCenter / lakeRadius;
		float exponent = 6.0;

		alpha = 1.0 - pow(t, exponent);
		alpha = clamp(alpha, 0.0, 1.0);

		float uv_scale = 4.0;
		vec2 uv = frag_in.textureCoordinates * uv_scale;
		float speed = 0.02;
        uv += vec2(speed * ub_data.time, speed * ub_data.time); // Scroll UVs
		vec3 normalTex = texture(normal_map, uv).rgb;
        normalTex = normalTex * 2.0 - 1.0;
		n = normalize(n + normalTex * 0.2);
    } else {
	    diffuseColor = texture(diffuse_texture, frag_in.textureCoordinates).rgb;
	}
	
	// Start with ambient illumination contribution:
	vec3 color = diffuseColor * ub_data.illumination[0];
	
	// Add directional light's contribution:
	color += phong(
		n, 
		-dl_data.direction.xyz, 
		-v, 
		dl_data.color.rgb * diffuseColor, ub_data.illumination[1], 
		dl_data.color.rgb,                ub_data.illumination[2], 
		ub_data.illumination[3], 
		false, vec3(1.0)
	);
			
	// Add point light's contribution
	color += phong(
		n, 
		pl_data.position.xyz - frag_in.positionWorld, 
		-v, 
		pl_data.color.rgb * diffuseColor, ub_data.illumination[1], 
		pl_data.color.rgb,                ub_data.illumination[2], 
		ub_data.illumination[3], 
		true, pl_data.attenuation.xyz
	);

	// Write color for the current fragment:
	out_color = vec4(color, alpha);

	if (ub_data.userInput[0] == 1) {
        out_color = vec4(n, alpha);
    }
	if (ub_data.userInput[1] == 1) {
        out_color = vec4(frag_in.textureCoordinates, 0, alpha);
    }
}
