#version 450
/*
 * Copyright 2023 TU Wien, Institute of Visual Computing & Human-Centered Technology.
 * This file is part of the GCG Lab Framework and must not be redistributed.
 *
 * Original version created by Lukas Gersthofer and Bernhard Steiner.
 * Vulkan edition created by Johannes Unterguggenberger (junt@cg.tuwien.ac.at).
 */

layout (location = 0) in vec3 in_position;
layout (location = 1) in vec3 in_normal;
layout (location = 2) in vec2 in_texture_coordinates;

layout (binding = 0) uniform UniformBuffer {
	vec4 color;
	mat4 modelMatrix;
	mat4 modelMatrixForNormals;
	mat4 viewProjMatrix;
	vec4 cameraPosition;
	vec4 illumination; // ka, kd, ks, alpha
	ivec4 userInput;
	
	float time;
    float impactX;
    float impactY;
    float pad3;
} ub_data;

layout (binding = 1) uniform DirectionalLight {
	vec4 color;
	vec4 direction;
} dl_data;

layout (binding = 2) uniform PointLight {
	vec4 color;
	vec4 position;
	vec4 attenuation;
} pl_data;

layout (location = 0) out VertexData {
	vec3 positionWorld;
	vec3 normalWorld;
	vec2 textureCoordinates;
} vert_out;

void main() {
    vec3 displacedPosition = in_position;
    vec3 displacedNormal = in_normal;
	
if (ub_data.userInput[2] == 1) {
        float amplitude = 0.03;
        float frequency = 45.0;
        float speed     = 0.2;
        float damping   = 10.0;

        float x = displacedPosition.x - ub_data.impactX;
        float z = displacedPosition.z - ub_data.impactY;
        float r = length(vec2(x, z));

        float wave = sin(frequency * (r - speed * ub_data.time));

        float amplitudeFalloff = amplitude / (1.0 + r * damping);
        amplitudeFalloff *= (1.0 - exp(-r * 10.0));

        // A small sink near the center
        float yFalloff = exp(-r * 5.0) / 15.0;

        float f = (wave * amplitudeFalloff) - yFalloff;
        displacedPosition.y += f;

        // Compute partial derivatives for normal
        // f(x,z) = wave(x,z,t)*amplitudeFalloff(r) - yFalloff(r)
        //
        //   dist(r) = sqrt(x^2 + z^2)
        //   wave    = sin( freq*(dist - speed*time) )
        //   amplitudeFalloff = A(r)
        //   yFalloff = Y(r)
        //   f(x,z) = wave * A(r) - Y(r)

        //    Derivative of wave wrt r
        //    wave = sin(...), where (...) = frequency*(r - speed*time)
        //    dwave/dr = cos(...)*frequency

        float A = frequency * (r - speed * ub_data.time);
        float dwave_dr = cos(A) * frequency;

        //    Derivative of amplitudeFalloff wrt r
        //    amplitudeFalloff = amplitudeFalloff(r).
        //    amplitudeFalloff(r) = amplitude*(1/(1 + r*damping))*(1 - exp(-r*10.0))
        //    break it in two factors: h(r) = 1/(1 + r*damping), q(r) = (1 - exp(-r*10))
        //    amplitudeFalloff = amplitude * h(r)*q(r)
        //
        //    d(amplitudeFalloff)/dr = amplitude * [h'(r)*q(r) + h(r)*q'(r)]
        //
        //    h(r) = 1/(1 + r*damping)
        //    h'(r) = -(damping)/(1 + r*damping)^2
        //
        //    q(r) = 1 - exp(-10*r)
        //    q'(r) = 10 * exp(-10*r)

        float h      = 1.0 / (1.0 + r*damping);
        float hPrime = -damping / ((1.0 + r*damping)*(1.0 + r*damping));

        float q      = 1.0 - exp(-10.0 * r);
        float qPrime = 10.0 * exp(-10.0 * r);

        float dAmpFalloff_dr = amplitude * (hPrime * q + h * qPrime);

        //    Derivative of yFalloff wrt r
        //    yFalloff = exp(-5*r)/15
        //    dyFalloff/dr = (1/15)*(-5)*exp(-5*r) = -5/15 * exp(-5*r)

        float dyFalloff_dr = -(1.0/3.0)*exp(-5.0*r); // -5/15 = -1/3

        //    f(x,z) = wave*A(r) - Y(r)
        //    df/dr = (dwave/dr)*A(r) + wave*dA(r)/dr - dY(r)/dr

        float df_dr = dwave_dr * amplitudeFalloff
                    + wave * dAmpFalloff_dr
                    - dyFalloff_dr;

        //    Convert df/dr to df/dx, df/dz
        //    r = sqrt(x^2 + z^2)
        //    dr/dx = x / r
        //    dr/dz = z / r

        float inv_r = (r < 1e-6) ? 0.0 : (1.0 / r);
        float dfdx = df_dr * (x * inv_r);
        float dfdz = df_dr * (z * inv_r);

        //    Create the local normal
        //    n_local = [ -df/dx, 1, -df/dz ]

        vec3 n_local = vec3(-dfdx, 1.0, -dfdz);

        displacedNormal = normalize(n_local);
    }
	
	vec4 position_world = ub_data.modelMatrix * vec4(displacedPosition, 1);
	vert_out.positionWorld = position_world.xyz;
	gl_Position = ub_data.viewProjMatrix * position_world;
	
	vert_out.normalWorld = mat3(ub_data.modelMatrixForNormals) * displacedNormal;
	vert_out.textureCoordinates = in_texture_coordinates;
}
