#version 460

layout(triangles, equal_spacing, ccw) in;

// === BINDINGS ===
layout (binding = 0) uniform UniformBuffer {
    vec4 color;
    mat4 modelMatrix;
    mat4 modelMatrixForNormals;
    mat4 viewProjMatrix;
    vec4 cameraPosition;
    vec4 illumination;
    ivec4 userInput;

    float time;
    float impactX;
    float impactY;
    float pad3;
} ub_data;

layout (binding = 1) uniform DirectionalLight {
    vec4 color;
    vec4 direction;
} dl_data;

layout (binding = 2) uniform PointLight {
    vec4 color;
    vec4 position;
    vec4 attenuation;
} pl_data;

layout (binding = 3) uniform sampler2D diffuse_texture;
layout (binding = 4) uniform sampler2D normal_map;


// === Input from TESC ===
struct VertexData
{
    vec3 posOS;      
    vec3 normalOS;   
    vec2 texcoords;
};

layout (location = 0) in VertexData tcs_out[];

// === Output to Fragment Shader ===
struct FragmentData {
    vec3 positionWorld;
    vec3 normalWorld;
    vec2 textureCoordinates;
};
layout (location = 0) out FragmentData te_out;

// Compute partial derivatives for wave displacement.
void computeWaveAndNormal(
    inout vec3 posOS,    
    inout vec3 normalOS,
    float x, float z
)
{
    if (ub_data.userInput[2] == 1)
    {
        float amplitude = 0.03;
        float frequency = 45.0;
        float speed     = 0.2;
        float damping   = 10.0;

        float dx = x - ub_data.impactX;
        float dz = z - ub_data.impactY;
        float r = length(vec2(dx, dz));

        float wave = sin(frequency * (r - speed * ub_data.time));

        float amplitudeFalloff = amplitude / (1.0 + r * damping);
        amplitudeFalloff *= (1.0 - exp(-r * 10.0));

        // A small sink near center
        float yFalloff = exp(-r * 5.0) / 15.0;

        float f = (wave * amplitudeFalloff) - yFalloff;
        posOS.y += f;

        // -------------- partial derivatives
        float A = frequency * (r - speed*ub_data.time);
        float dwave_dr = cos(A)*frequency;

        float h = 1.0/(1.0 + r*damping);
        float hPrime = -damping / pow(1.0 + r*damping, 2.0);
        float q = 1.0 - exp(-10.0*r);
        float qPrime = 10.0*exp(-10.0*r);
        float dAmpFalloff_dr = amplitude*(hPrime*q + h*qPrime);

        float dyFalloff_dr = -(1.0/3.0)*exp(-5.0*r);

        float df_dr = (dwave_dr*amplitudeFalloff) + (wave*dAmpFalloff_dr) - dyFalloff_dr;

        float inv_r = (r<1e-6)?0.0:(1.0/r);
        float dfdx = df_dr*(dx*inv_r);
        float dfdz = df_dr*(dz*inv_r);

        // local normal is [-dfdx, 1, -dfdz]
        vec3 n_local = vec3(-dfdx, 1.0, -dfdz);
        normalOS = normalize(n_local);
    }
    else
    {
        // if wave is off, normal remains as is
    }
}

void main()
{
    // Barycentric coords
    float u = gl_TessCoord.x;
    float v = gl_TessCoord.y;
    float w = gl_TessCoord.z;

    // Interpolate corners
    vec3 posOS = u * tcs_out[0].posOS + v * tcs_out[1].posOS + w * tcs_out[2].posOS;
    vec3 normalOS = normalize(u * tcs_out[0].normalOS + v * tcs_out[1].normalOS + w * tcs_out[2].normalOS);
    vec2 tex = u * tcs_out[0].texcoords + v * tcs_out[1].texcoords + w * tcs_out[2].texcoords;

    computeWaveAndNormal(posOS, normalOS, posOS.x, posOS.z);

    // Transform to world space
    vec4 posWorld4 = ub_data.modelMatrix * vec4(posOS,1);
    te_out.positionWorld = posWorld4.xyz;
    vec3 nWorld = mat3(ub_data.modelMatrixForNormals)*normalOS;
    te_out.normalWorld = normalize(nWorld);
    te_out.textureCoordinates = tex;
    gl_Position = ub_data.viewProjMatrix * posWorld4;
}
 