/*
 * Copyright 2023 TU Wien, Institute of Visual Computing & Human-Centered Technology.
 * This file is part of the GCG Lab Framework and must not be redistributed.
 *
 * Original version created by Lukas Gersthofer and Bernhard Steiner.
 * Vulkan edition created by Johannes Unterguggenberger (junt@cg.tuwien.ac.at).
 */

#include "PathUtils.h"
#include "Utils.h"

#include <vulkan/vulkan.h>

#include <vector>
#include <array>
#include <fstream>
#include <string>

#include <sstream>
#include <cassert>
#include <iostream>
#include "Camera.h"
#include "ObjLoader.h"

#include <tiny_obj_loader.h>

#define USE_GLSLANG

#ifdef USE_SHADERC
#   include <shaderc/shaderc.hpp>
#endif
#ifdef USE_GLSLANG
#   include <glslang/Include/glslang_c_interface.h>
#endif

#undef min
#undef max

constexpr char WELCOME_MSG[] = ":::::: Nature's Seclude ::::::";
constexpr char WINDOW_TITLE[] = "RTG 2024";

constexpr float BACKGROUND_R = 0.14f;
constexpr float BACKGROUND_G = 0.4f;
constexpr float BACKGROUND_B = 0.37f;

constexpr glm::vec4 DIRLIGHT_COLOR = glm::vec4(0.85f, 0.85f, 0.85f, 0.0f);
constexpr glm::vec4 DIRLIGHT_DIR = glm::vec4(0.5f, -1.0f, -0.0f, 0.0f);
constexpr glm::vec4 POINTLIGHT_COLOR = glm::vec4(0.0f, 0.0f, 0.0f, 0.0f);
constexpr glm::vec4 POINTLIGHT_POS = glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);
constexpr glm::vec4 POINTLIGHT_ATTENUATION = glm::vec4(2.0f, 0.0f, 0.0f, 0.0f);

constexpr float KA = 0.1f;
constexpr float KD = 0.9f;
constexpr float KS = 0.05f;

constexpr float KA_water = 0.1f;
constexpr float KD_water = 0.9f;
constexpr float KS_water = 0.3f;

constexpr float ALPHA = 10.0f;

static bool manualCameraMode = false;


/* --------------------------------------------- */
// Helper Function Declarations
/* --------------------------------------------- */

/*!
 *	Add extension_name to the target vector ref_vector if the extension is supported on this system.
 *	@param	extension_name		The instance extension that shall be added to ref_vector
 *	@param	ref_vector			Reference to the vector that the instance extension shall be added to, if supported
 */
void addInstanceExtensionToVectorIfSupported(const char* extension_name, std::vector<const char*>& ref_vector);

/*!
 *	Add validation_layer_name to the target vector ref_vector if the validation layer is supported on this system:
 *	@param	validation_layer_name	The validation layer name that shall be added to ref_vector
 *	@param	ref_vector				Reference to the vector that the validation layer name shall be added to, if
 *supported
 */
void addValidationLayerNameToVectorIfSupported(const char* validation_layer_name, std::vector<const char*>& ref_vector);

/*!
 *	Add extension_name to the target vector ref_vector if the extension is supported by the given physical device.
 *	@param	extension_name		The device extension that shall be added to ref_vector
 *	@param	physical_device		The physical device handle which must support the given extension
 *	@param	ref_vector			Reference to the vector that the device extension shall be added to, if supported
 */
void addDeviceExtensionToVectorIfSupported(const char* extension_name, VkPhysicalDevice physical_device, std::vector<const char*>& ref_vector);
/*!
 *	From the given list of physical devices, select the first one that satisfies all requirements.
 *	@param		physical_devices		A pointer which points to contiguous memory of #physical_device_count sequentially
                                        stored VkPhysicalDevice handles is expected. The handles can (or should) be those
 *										that are returned from vkEnumeratePhysicalDevices.
 *	@param		physical_device_count	The number of consecutive physical device handles there are at the memory location
 *										that is pointed to by the physical_devices parameter.
 *	@param		surface					A valid VkSurfaceKHR handle, which is used to determine if a certain
 *										physical device supports presenting images to the given surface.
 *	@return		The index of the physical device that satisfies all requirements is returned.
 */
uint32_t selectPhysicalDeviceIndex(const VkPhysicalDevice* physical_devices, uint32_t physical_device_count, VkSurfaceKHR surface);

/*!
 *	From the given list of physical devices, select the first one that satisfies all requirements.
 *	@param		physical_devices	A vector containing all available VkPhysicalDevice handles, like those
 *									that are returned from vkEnumeratePhysicalDevices.
 *	@param		surface				A valid VkSurfaceKHR handle, which is used to determine if a certain
 *									physical device supports presenting images to the given surface.
 *	@return		The index of the physical device that satisfies all requirements is returned.
 */
uint32_t selectPhysicalDeviceIndex(const std::vector<VkPhysicalDevice>& physical_devices, VkSurfaceKHR surface);

/*!
 *	Based on the given physical device and the surface, select a queue family which supports both,
 *	graphics and presentation to the given surface. Return the INDEX of an appropriate queue family!
 *	@return		The index of a queue family which supports the required features shall be returned.
 */
uint32_t selectQueueFamilyIndex(VkPhysicalDevice physical_device, VkSurfaceKHR surface);

/*!
 *	Based on the given physical device and the surface, a the physical device's surface capabilites are read and returned.
 *	@return		VkSurfaceCapabilitiesKHR data
 */
VkSurfaceCapabilitiesKHR getPhysicalDeviceSurfaceCapabilities(VkPhysicalDevice physical_device, VkSurfaceKHR surface);

/*!
 *	Based on the given physical device and the surface, a supported surface image format
 *	which can be used for the framebuffer's attachment formats is searched and returned.
 *	@return		A supported format is returned.
 */
VkSurfaceFormatKHR getSurfaceImageFormat(VkPhysicalDevice physical_device, VkSurfaceKHR surface);

/*!
 *	Based on the given physical device and the surface, return its surface transform flag.
 *	This can be used to set the swap chain to the same configuration as the surface's current transform.
 *	@return		The surface capabilities' currentTransform value is returned, which is suitable for swap chain config.
 */
VkSurfaceTransformFlagBitsKHR getSurfaceTransform(VkPhysicalDevice physical_device, VkSurfaceKHR surface);

void writeParticleDescriptorSet(VkDevice device, VkDescriptorSet set,
    VkBuffer uniform_buffer, VkBuffer storage_buffer);

/*!
 *	It matches the definition and sizes of the corresponding GPU-side struct exactly, which is used in shaders.
 */
struct UniformBuffer {
    /*! A color values, stored as four floats, to properly align to 16 byte boundaries. */
    glm::vec4 color;

    /*! Storage for the model matrix, consisting of 16 float values (inherently aligned to 16 bytes) */
    glm::mat4 modelMatrix;

    /*! Storage for the model matrix suitable for normals transformation, consisting of 16 float values (inherently aligned to 16 bytes) */
    glm::mat4 modelMatrixForNormals;

    /*! Storage for the view-projection matrix, consisting of 16 float values (inherently aligned to 16 bytes) */
    glm::mat4 viewProjMatrix;

    /*! Storage for the camera's world space position (aligned to 16 bytes) */
    glm::vec4 cameraPosition;

    /*! Illumination properties ka, kd, ks, alpha (in that order)
     *	First three are material coefficients, the last one is specular alpha. */
    glm::vec4 materialProperties;

    /*! stores user input as magic numbers */
    glm::ivec4 userInput;

    /*! stores time value */
    float time;
    float impactX;
    float impactY;
    float pad;
};

/*!
 *	This struct contains the data of a directional light.
 *	It matches the definition and sizes of the corresponding GPU-side struct exactly, which is used in shaders.
 */
struct DirectionalLight {
    /*! Light color of this light source */
    glm::vec4 color;

    /*! Light direction of this directional light source */
    glm::vec4 direction;
};

/*!
 *	This struct contains the data of a point light.
 *	It matches the definition and sizes of the corresponding GPU-side struct exactly, which is used in shaders.
 */
struct PointLight {
    /*! Light color of this light source */
    glm::vec4 color;

    /*! Position of this light source in world space */
    glm::vec4 position;

    /*! Attenuation properties of this light source */
    glm::vec4 attenuation;
};

/*!
 *	Allocates a new descriptor set of the given layout from the given descriptor pool.
 *	It is not required to cleanup the returned descriptor set explicitly, it will be cleaned up when the descriptor pool is destroyed.
 *	@param	device					Valid handle to the logical device
 *	@param	descriptor_pool			Valid handle to a descriptor pool
 *	@param	descriptor_set_layout	Valid handle to a descriptor set layout
 */
VkDescriptorSet allocDescriptorSet(VkDevice device, VkDescriptorPool descriptor_pool, VkDescriptorSetLayout descriptor_set_layout);

/*!
 *	Writes the descriptor information to a given descriptor set which describes one uniform buffer at binding = 0.
 *	The given descriptor set must have been created from a descriptor set layout according to this structure.
 *	@param	device					Valid handle to the logical device
 *	@param	descriptor_set			Valid handle to a descriptor set, which concrete descriptor information will be written to.
 *	@param	uniform_buffer			A descriptor for this uniform buffer will be written to binding = 0
 */
void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer uniform_buffer);

/*!
 *	Writes the descriptor information to a given descriptor set which describes one uniform buffer at binding = 0,
 *	another uniform buffer at binding = 1, and yet another uniform buffer at binding = 2.
 *	The given descriptor set must have been created from a descriptor set layout according to this structure.
 *	@param	device					Valid handle to the logical device
 *	@param	descriptor_set			Valid handle to a descriptor set, which concrete descriptor information will be written to.
 *	@param	object_data				A descriptor for this uniform buffer will be written to binding = 0
 *	@param	directional_light_data	A descriptor for this uniform buffer will be written to binding = 1
 *	@param	point_light_data		A descriptor for this uniform buffer will be written to binding = 2
 */
void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer object_data, VkBuffer directional_light_data, VkBuffer point_light_data);

/*!
 *	Writes the descriptor information to a given descriptor set which describes one uniform buffer at binding = 0,
 *	another uniform buffer at binding = 1, and yet another uniform buffer at binding = 2, furthermore,
 *	a combined image sampler descriptor is written for binding = 3
 *	The given descriptor set must have been created from a descriptor set layout according to this structure.
 *	@param	device					Valid handle to the logical device
 *	@param	descriptor_set			Valid handle to a descriptor set, which concrete descriptor information will be written to.
 *	@param	object_data				A descriptor for this uniform buffer will be written to binding = 0
 *	@param	directional_light_data	A descriptor for this uniform buffer will be written to binding = 1
 *	@param	point_light_data		A descriptor for this uniform buffer will be written to binding = 2
 *	@param	image_view				A combined descriptor together with sampler will be written to binding = 3
 *	@param	sampler					A combined descriptor together with image_view will be written to binding = 3
 */
void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer object_data, VkBuffer directional_light_data, VkBuffer point_light_data,
    VkImageView image_view, VkSampler sampler);

/*!
 *	Writes the descriptor information to a given descriptor set which describes one uniform buffer at binding = 0,
 *	another uniform buffer at binding = 1, and yet another uniform buffer at binding = 2, furthermore,
 *	a combined image sampler descriptor is written for binding = 3
 *	The given descriptor set must have been created from a descriptor set layout according to this structure.
 *	@param	device					Valid handle to the logical device
 *	@param	descriptor_set			Valid handle to a descriptor set, which concrete descriptor information will be written to.
 *	@param	object_data				A descriptor for this uniform buffer will be written to binding = 0
 *	@param	directional_light_data	A descriptor for this uniform buffer will be written to binding = 1
 *	@param	point_light_data		A descriptor for this uniform buffer will be written to binding = 2
 *	@param	image_view				A combined descriptor together with sampler will be written to binding = 3
 *	@param	sampler					A combined descriptor together with image_view will be written to binding = 3
 *	@param	normalMapView			A combined descriptor together with normalMapSampler will be written to binding = 4
 *	@param	normalMapSampler		A combined descriptor together with normalMapView will be written to binding = 4
 */
void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer object_data, VkBuffer directional_light_data, VkBuffer point_light_data,
    VkImageView image_view, VkSampler sampler, VkImageView normalMapView, VkSampler normalMapSampler);

/*!
 *	This callback function gets invoked by GLFW whenever a GLFW error occured.
 */
void errorCallbackFromGlfw(int error, const char* description) { std::cout << "GLFW error " << error << ": " << description << std::endl; }

static bool g_dragging = false;
static bool g_strafing = false;
static float g_zoom = 5.0f;

static float g_time = 0.0f;

/*!
 *	This callback function gets invoked by GLFW during glfwPollEvents() if there was
 *	mouse button input that can be processed by our application.
 */
void mouseButtonCallbackFromGlfw(GLFWwindow* glfw_window, int button, int action, int mods) {
    if (button == GLFW_MOUSE_BUTTON_LEFT && action == GLFW_PRESS) {
        g_dragging = true;
    } else if (button == GLFW_MOUSE_BUTTON_LEFT && action == GLFW_RELEASE) {
        g_dragging = false;
    } else if (button == GLFW_MOUSE_BUTTON_RIGHT && action == GLFW_PRESS) {
        g_strafing = true;
    } else if (button == GLFW_MOUSE_BUTTON_RIGHT && action == GLFW_RELEASE) {
        g_strafing = false;
    }
}

/*!
 *	This callback function gets invoked by GLFW during glfwPollEvents() if there was
 *	mouse scroll input that can be processed by our application.
 */
void scrollCallbackFromGlfw(GLFWwindow* glfw_window, double xoffset, double yoffset) { g_zoom -= static_cast<float>(yoffset) * 0.5f; }


/*!
 *	0 ... fill polygons
 *	1 ... wireframe mode
 */
static int g_polygon_mode_index = 0;

/*!
 *	0 ... no culling
 *	1 ... cull back faces
 *	2 ... cull front faces
 */
static int g_culling_index = 0;

/*!
 *	Bind the given descriptor se to use the material it represents for subsequent draw calls
 *	with the given pipeline, and render the given geometry (using its vertex and index buffers).
 *	Record everything into the current command buffer as provided by the framework.
 *	@param	pipeline		Valid handle to a given pipeline which shall be used for drawing.
 *	@param	geometry		Reference to a geometry object containing the buffers to be used for drawing.
 *	@param	material		Valid handle to a descriptor set that refers to resources that contain material properties.
 *	@param	num_instances	How many instances to draw of the given geometry. Default = one single instance.
 */
void drawGeometryWithMaterial(VkPipeline pipeline, const Geometry& geometry, VkDescriptorSet material, uint32_t num_instances, VkCommandBuffer command_buffer);

static bool g_draw_normals = false;
static bool g_draw_texcoords = false;

/*!
 *	A flag that will be set during initialization code.
 *	If set to true, it will indicate that Synchronization2 is supported and can be used.
 */
static bool g_synchronization2_supported = false;

/*!
 *	Function pointer to the implementation of vkCmdPipelineBarrier2KHR() which is not
 *	loaded by default since it is an extension. We have to get the pointer manually,
 *	and we do that only in the case where Synchronization2 is supported.
 */
PFN_vkCmdPipelineBarrier2KHR g_vkCmdPipelineBarrier2KHR;

/*!
 *	Creates an image view for the given image and returns its handle.
 *	The returned handle is supposed to be destroyed with vkDestroyImageView
 */
VkImageView createImageViewForImage(VkDevice device, VkImage image, VkImageViewType view_type, VkFormat format);

/*!
 *	Struct used as the return value for loadImage
 */
struct ImageAndView {
    /*! Contains the created image handle */
    VkImage image;

    /*! Contains the created image view handle */
    VkImageView view;
};

/*!
 *	Copies the data from file into a VkBuffer, and then copies it from that buffer into the given image.
 *	Waits until the operation has finished on the GPU.
 */
ImageAndView loadImage(VkDevice device, VkQueue queue, VkCommandPool command_pool, std::string image_file_name);

struct CustomRenderPassInfo {
    VkRenderPass renderPass = VK_NULL_HANDLE;
    std::vector<VkFramebuffer> framebuffers;
};

CustomRenderPassInfo createCustomRenderPassAndFramebuffers(
    VkDevice device,
    const VklSwapchainConfig& swapchain_config
);

VkImageView createImageViewForSwapchainImage(
    VkDevice device,
    const VklSwapchainImageDetails& imgDetails
);

float randf(int seed) {
    float r = std::sin(static_cast<float>(seed) * 12.9898f) * 43758.5453f;
    return r - std::floor(r);
}

static std::string readFileAsString(const std::string& path)
{
    std::ifstream ifs(path, std::ios::in | std::ios::binary);
    if (!ifs.good()) {
        throw std::runtime_error("Failed to open file [" + path + "]");
    }
    std::string contents((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));
    return contents;
}

struct Particle {
    glm::vec4 position;
    glm::vec4 velocity;
    float killPlaneHeight; // Particle should die when it falls below that height (water surface)
    float pad[3];
};

constexpr uint32_t NUM_PARTICLES = 100000;

#ifdef USE_GLSLANG

glslang_resource_t get_default_resource_main() {
    glslang_resource_t r = {
        /* .MaxLights = */ 32,
        /* .MaxClipPlanes = */ 6,
        /* .MaxTextureUnits = */ 32,
        /* .MaxTextureCoords = */ 32,
        /* .MaxVertexAttribs = */ 64,
        /* .MaxVertexUniformComponents = */ 4096,
        /* .MaxVaryingFloats = */ 64,
        /* .MaxVertexTextureImageUnits = */ 32,
        /* .MaxCombinedTextureImageUnits = */ 80,
        /* .MaxTextureImageUnits = */ 32,
        /* .MaxFragmentUniformComponents = */ 4096,
        /* .MaxDrawBuffers = */ 32,
        /* .MaxVertexUniformVectors = */ 128,
        /* .MaxVaryingVectors = */ 8,
        /* .MaxFragmentUniformVectors = */ 16,
        /* .MaxVertexOutputVectors = */ 16,
        /* .MaxFragmentInputVectors = */ 15,
        /* .MinProgramTexelOffset = */ -8,
        /* .MaxProgramTexelOffset = */ 7,
        /* .MaxClipDistances = */ 8,
        /* .MaxComputeWorkGroupCountX = */ 65535,
        /* .MaxComputeWorkGroupCountY = */ 65535,
        /* .MaxComputeWorkGroupCountZ = */ 65535,
        /* .MaxComputeWorkGroupSizeX = */ 1024,
        /* .MaxComputeWorkGroupSizeY = */ 1024,
        /* .MaxComputeWorkGroupSizeZ = */ 64,
        /* .MaxComputeUniformComponents = */ 1024,
        /* .MaxComputeTextureImageUnits = */ 16,
        /* .MaxComputeImageUniforms = */ 8,
        /* .MaxComputeAtomicCounters = */ 8,
        /* .MaxComputeAtomicCounterBuffers = */ 1,
        /* .MaxVaryingComponents = */ 60,
        /* .MaxVertexOutputComponents = */ 64,
        /* .MaxGeometryInputComponents = */ 64,
        /* .MaxGeometryOutputComponents = */ 128,
        /* .MaxFragmentInputComponents = */ 128,
        /* .MaxImageUnits = */ 8,
        /* .MaxCombinedImageUnitsAndFragmentOutputs = */ 8,
        /* .MaxCombinedShaderOutputResources = */ 8,
        /* .MaxImageSamples = */ 0,
        /* .MaxVertexImageUniforms = */ 0,
        /* .MaxTessControlImageUniforms = */ 0,
        /* .MaxTessEvaluationImageUniforms = */ 0,
        /* .MaxGeometryImageUniforms = */ 0,
        /* .MaxFragmentImageUniforms = */ 8,
        /* .MaxCombinedImageUniforms = */ 8,
        /* .MaxGeometryTextureImageUnits = */ 16,
        /* .MaxGeometryOutputVertices = */ 256,
        /* .MaxGeometryTotalOutputComponents = */ 1024,
        /* .MaxGeometryUniformComponents = */ 1024,
        /* .MaxGeometryVaryingComponents = */ 64,
        /* .MaxTessControlInputComponents = */ 128,
        /* .MaxTessControlOutputComponents = */ 128,
        /* .MaxTessControlTextureImageUnits = */ 16,
        /* .MaxTessControlUniformComponents = */ 1024,
        /* .MaxTessControlTotalOutputComponents = */ 4096,
        /* .MaxTessEvaluationInputComponents = */ 128,
        /* .MaxTessEvaluationOutputComponents = */ 128,
        /* .MaxTessEvaluationTextureImageUnits = */ 16,
        /* .MaxTessEvaluationUniformComponents = */ 1024,
        /* .MaxTessPatchComponents = */ 120,
        /* .MaxPatchVertices = */ 32,
        /* .MaxTessGenLevel = */ 64,
        /* .MaxViewports = */ 16,
        /* .MaxVertexAtomicCounters = */ 0,
        /* .MaxTessControlAtomicCounters = */ 0,
        /* .MaxTessEvaluationAtomicCounters = */ 0,
        /* .MaxGeometryAtomicCounters = */ 0,
        /* .MaxFragmentAtomicCounters = */ 8,
        /* .MaxCombinedAtomicCounters = */ 8,
        /* .MaxAtomicCounterBindings = */ 1,
        /* .MaxVertexAtomicCounterBuffers = */ 0,
        /* .MaxTessControlAtomicCounterBuffers = */ 0,
        /* .MaxTessEvaluationAtomicCounterBuffers = */ 0,
        /* .MaxGeometryAtomicCounterBuffers = */ 0,
        /* .MaxFragmentAtomicCounterBuffers = */ 1,
        /* .MaxCombinedAtomicCounterBuffers = */ 1,
        /* .MaxAtomicCounterBufferSize = */ 16384,
        /* .MaxTransformFeedbackBuffers = */ 4,
        /* .MaxTransformFeedbackInterleavedComponents = */ 64,
        /* .MaxCullDistances = */ 8,
        /* .MaxCombinedClipAndCullDistances = */ 8,
        /* .MaxSamples = */ 4,
        /* .maxMeshOutputVerticesNV = */ 256,
        /* .maxMeshOutputPrimitivesNV = */ 512,
        /* .maxMeshWorkGroupSizeX_NV = */ 32,
        /* .maxMeshWorkGroupSizeY_NV = */ 1,
        /* .maxMeshWorkGroupSizeZ_NV = */ 1,
        /* .maxTaskWorkGroupSizeX_NV = */ 32,
        /* .maxTaskWorkGroupSizeY_NV = */ 1,
        /* .maxTaskWorkGroupSizeZ_NV = */ 1,
        /* .maxMeshViewCountNV = */ 4,
        /* .maxDualSourceDrawBuffersEXT = */ 1,
        /* .limits = */ 
            ///* .nonInductiveForLoops = */ 1,
            ///* .whileLoops = */ 1,
            ///* .doWhileLoops = */ 1,
            ///* .generalUniformIndexing = */ 1,
            ///* .generalAttributeMatrixVectorIndexing = */ 1,
            ///* .generalVaryingIndexing = */ 1,
            ///* .generalSamplerIndexing = */ 1,
            ///* .generalVariableIndexing = */ 1,
            ///* .generalConstantMatrixVectorIndexing = */ 1,
            
    };

    r.limits.non_inductive_for_loops = 1;
    r.limits.while_loops = 1;
    r.limits.do_while_loops = 1;
    r.limits.general_uniform_indexing = 1;
    r.limits.general_attribute_matrix_vector_indexing = 1;
    r.limits.general_varying_indexing = 1;
    r.limits.general_constant_matrix_vector_indexing = 1;
    r.limits.general_sampler_indexing = 1;
    r.limits.general_variable_indexing = 1;
    return r;
}
#endif

std::vector<VkImageView> imageViews;

static std::vector<uint32_t> compileGlslToSpirv(
    const std::string& glslSource,
    const std::string& debugFilename,
    VkShaderStageFlagBits stage
)
{
    // Stage mapping: map Vulkan's stage to either shaderc or glslang enumerations
#ifdef USE_SHADERC
    shaderc_shader_kind shadercKind;
    switch (stage) {
    case VK_SHADER_STAGE_VERTEX_BIT:
        shadercKind = shaderc_vertex_shader;
        break;
    case VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT:
        shadercKind = shaderc_tess_control_shader;
        break;
    case VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT:
        shadercKind = shaderc_tess_evaluation_shader;
        break;
    case VK_SHADER_STAGE_GEOMETRY_BIT:
        shadercKind = shaderc_geometry_shader;
        break;
    case VK_SHADER_STAGE_FRAGMENT_BIT:
        shadercKind = shaderc_fragment_shader;
        break;
    case VK_SHADER_STAGE_COMPUTE_BIT:
        shadercKind = shaderc_compute_shader;
        break;
    default:
        throw std::runtime_error("Unsupported stage in compileGlslToSpirv (shaderc).");
    }

    shaderc::Compiler compiler;
    shaderc::CompileOptions options;

    shaderc::SpvCompilationResult spirvModule =
        compiler.CompileGlslToSpv(glslSource, shadercKind, debugFilename.c_str(), options);

    if (spirvModule.GetCompilationStatus() != shaderc_compilation_status_success) {
        std::cerr << "ERROR: compileGlslToSpirv [" << debugFilename << "] "
            << spirvModule.GetErrorMessage() << std::endl;
        throw std::runtime_error("Failed to compile shader [" + debugFilename + "]");
    }

    return std::vector<uint32_t>(spirvModule.cbegin(), spirvModule.cend());

#elif defined(USE_GLSLANG)
    glslang_stage_t glslangStage;
    switch (stage) {
    case VK_SHADER_STAGE_VERTEX_BIT:
        glslangStage = GLSLANG_STAGE_VERTEX;
        break;
    case VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT:
        glslangStage = GLSLANG_STAGE_TESSCONTROL;
        break;
    case VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT:
        glslangStage = GLSLANG_STAGE_TESSEVALUATION;
        break;
    case VK_SHADER_STAGE_GEOMETRY_BIT:
        glslangStage = GLSLANG_STAGE_GEOMETRY;
        break;
    case VK_SHADER_STAGE_FRAGMENT_BIT:
        glslangStage = GLSLANG_STAGE_FRAGMENT;
        break;
    case VK_SHADER_STAGE_COMPUTE_BIT:
        glslangStage = GLSLANG_STAGE_COMPUTE;
        break;
    default:
        throw std::runtime_error("Unsupported stage in compileGlslToSpirv (glslang).");
    }

    std::vector<uint32_t> resultingSpirv;

    const char* shaderCode = glslSource.c_str();

    static const auto defaultResources = get_default_resource_main();

    glslang_input_t input = {};
    input.language = GLSLANG_SOURCE_GLSL;
    input.stage = glslangStage;
    input.client = GLSLANG_CLIENT_VULKAN;
    input.client_version = GLSLANG_TARGET_VULKAN_1_1;
    input.target_language = GLSLANG_TARGET_SPV;
    input.target_language_version = GLSLANG_TARGET_SPV_1_3;
    input.code = shaderCode;
    input.default_version = 100;
    input.default_profile = GLSLANG_NO_PROFILE;
    input.force_default_version_and_profile = false;
    input.forward_compatible = false;
    input.messages = GLSLANG_MSG_DEFAULT_BIT;
    input.resource = &defaultResources;

    glslang_shader_t* shader = glslang_shader_create(&input);

    if (!glslang_shader_preprocess(shader, &input)) {
        std::cerr << "ERROR: Failed to preprocess shader [" << debugFilename << "]: "
            << glslang_shader_get_info_log(shader) << std::endl;
        throw std::runtime_error("glslang preprocessing failed for " + debugFilename);
    }

    if (!glslang_shader_parse(shader, &input)) {
        std::cerr << "ERROR: Failed to parse shader [" << debugFilename << "]: "
            << glslang_shader_get_info_log(shader) << std::endl;
        throw std::runtime_error("glslang parsing failed for " + debugFilename);
    }

    glslang_program_t* program = glslang_program_create();
    glslang_program_add_shader(program, shader);

    if (!glslang_program_link(program, GLSLANG_MSG_SPV_RULES_BIT | GLSLANG_MSG_VULKAN_RULES_BIT)) {
        std::cerr << "ERROR: Failed to link shader [" << debugFilename << "]: "
            << glslang_shader_get_info_log(shader) << std::endl;
        throw std::runtime_error("glslang linking failed for " + debugFilename);
    }

    glslang_program_SPIRV_generate(program, glslangStage);

    if (glslang_program_SPIRV_get_messages(program)) {
        std::cout << "INFO: Shader compilation messages for [" << debugFilename << "]: "
            << glslang_program_SPIRV_get_messages(program) << std::endl;
    }

    auto* spirvDataPtr = glslang_program_SPIRV_get_ptr(program);
    const auto spirvNumWords = glslang_program_SPIRV_get_size(program);
    resultingSpirv.insert(resultingSpirv.end(), spirvDataPtr, spirvDataPtr + spirvNumWords);

    glslang_program_delete(program);
    glslang_shader_delete(shader);

    return resultingSpirv;
#else
    std::cout << "else" << std::endl;
    throw std::runtime_error("No shaderc or glslang compile path defined!");
#endif
}

VkShaderModule createShaderModule(VkDevice device, const std::string& glslFilePath, VkShaderStageFlagBits stage)
{
    std::string glslText = readFileAsString(glslFilePath);
    std::vector<uint32_t> spirvCode = compileGlslToSpirv(glslText, glslFilePath, stage);

    std::cout << glslFilePath << std::endl;

    VkShaderModuleCreateInfo smCI{};
    smCI.sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
    smCI.codeSize = spirvCode.size() * sizeof(uint32_t);
    smCI.pCode = spirvCode.data();

    VkShaderModule shaderModule = VK_NULL_HANDLE;
    VkResult res = vkCreateShaderModule(device, &smCI, nullptr, &shaderModule);
    if (res != VK_SUCCESS) {
        throw std::runtime_error("Failed to create shader module for " + glslFilePath);
    }

    return shaderModule;
}

std::unordered_map<VkPipeline, VkPipelineLayout> g_PipelineLayoutMap;


/* --------------------------------------------- */
// Main
/* --------------------------------------------- */

int main(int argc, char** argv) {
    VKL_LOG(WELCOME_MSG);

    CMDLineArgs cmdline_args;
    gcgParseArgs(cmdline_args, argc, argv);

    // Load settings from file
    int window_width = 1920;
    int window_height = 1080;
    bool fullscreen = false;
    std::string window_title = "Nature's Seclude";
    INIReader window_reader("assets/settings/window.ini");

    std::string init_camera_filepath = "assets/settings/camera_front.ini";
    if (cmdline_args.init_camera) {
        init_camera_filepath = cmdline_args.init_camera_filepath;
    }
    INIReader camera_reader(init_camera_filepath);

    window_width = window_reader.GetInteger("window", "width", 1920);
    window_height = window_reader.GetInteger("window", "height", 1080);
    fullscreen = true;
    window_title = window_reader.Get("window", "title", WINDOW_TITLE);
    float camera_yaw = static_cast<float>(camera_reader.GetReal("camera", "yaw", 0.0f));
    float camera_pitch = static_cast<float>(camera_reader.GetReal("camera", "pitch", 0.0f));

    camera_pitch += 0.5;
    camera_yaw += 2.5;

    float field_of_view = static_cast<float>(camera_reader.GetReal("camera", "fov", 60.0f));
    float near_plane_distance = static_cast<float>(camera_reader.GetReal("camera", "near", 0.1f));
    float far_plane_distance = static_cast<float>(camera_reader.GetReal("camera", "far", 100.0f));
    float aspect_ratio = static_cast<float>(window_width) / static_cast<float>(window_height);

    std::string init_renderer_filepath = "assets/settings/renderer_standard.ini";
    if (cmdline_args.init_renderer) {
        init_renderer_filepath = cmdline_args.init_renderer_filepath;
    }
    INIReader renderer_reader(init_renderer_filepath);

    bool as_wireframe = renderer_reader.GetBoolean("renderer", "wireframe", false);
    if (as_wireframe) {
        g_polygon_mode_index = 1;
    }
    bool with_backface_culling = renderer_reader.GetBoolean("renderer", "backface_culling", false);
    if (with_backface_culling) {
        g_culling_index = 1;
    }
    g_draw_normals = renderer_reader.GetBoolean("renderer", "normals", false);
    g_draw_texcoords = renderer_reader.GetBoolean("renderer", "texcoords", false);
    bool depthtest = renderer_reader.GetBoolean("renderer", "depthtest", true);

    // Install a callback function for GLFW errors
    glfwSetErrorCallback(errorCallbackFromGlfw);

    // Initialize GLFW and create a window
    if (!glfwInit()) {
        VKL_EXIT_WITH_ERROR("Failed to initialize GLFW");
    }

    glfwWindowHint(GLFW_CLIENT_API, GLFW_NO_API); // No need to create an OpenGL context
    glfwWindowHint(GLFW_RESIZABLE, GLFW_FALSE);

    // Use a monitor for fullscreen mode if requested
    GLFWmonitor* monitor = nullptr;
    if (fullscreen) {
        monitor = glfwGetPrimaryMonitor();
    }

    GLFWwindow* window = glfwCreateWindow(window_width, window_height, window_title.c_str(), monitor, nullptr);
    if (!window) {
        VKL_LOG("Failed to create GLFW window.");
        VKL_EXIT_WITH_ERROR("No GLFW window created.");
    }

    VkResult result;
    VkInstance vk_instance = VK_NULL_HANDLE;
    VkSurfaceKHR vk_surface = VK_NULL_HANDLE;
    VkPhysicalDevice vk_physical_device = VK_NULL_HANDLE;
    VkDevice vk_device = VK_NULL_HANDLE;
    VkQueue vk_queue = VK_NULL_HANDLE;
    VkSwapchainKHR vk_swapchain = VK_NULL_HANDLE;

    // Create a Vulkan instance
    VkApplicationInfo application_info = {};
    application_info.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
    application_info.pEngineName = "GCG_VK_Library";
    application_info.engineVersion = VK_MAKE_API_VERSION(0, 2023, 9, 1);
    application_info.pApplicationName = "GCG_VK_Solution";
    application_info.applicationVersion = VK_MAKE_API_VERSION(0, 2023, 9, 19);
    application_info.apiVersion = VK_API_VERSION_1_1;

    VkInstanceCreateInfo instance_create_info = {};
    instance_create_info.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
    instance_create_info.pApplicationInfo = &application_info;

    // Gather required instance extensions
    std::vector<const char*> instance_extensions;

    // Extensions required by GLFW
    uint32_t glfw_instance_extensions_count;
    const char** glfw_instance_extensions_names = glfwGetRequiredInstanceExtensions(&glfw_instance_extensions_count);
    for (uint32_t i = 0; i < glfw_instance_extensions_count; ++i) {
        addInstanceExtensionToVectorIfSupported(glfw_instance_extensions_names[i], instance_extensions);
    }

    // Extensions required by the framework
    uint32_t framework_instance_extensions_count;
    const char** framework_instance_extensions_names = vklGetRequiredInstanceExtensions(&framework_instance_extensions_count);
    for (uint32_t i = 0; i < framework_instance_extensions_count; ++i) {
        addInstanceExtensionToVectorIfSupported(framework_instance_extensions_names[i], instance_extensions);
    }
#ifdef __APPLE__
    addInstanceExtensionToVectorIfSupported(VK_KHR_PORTABILITY_ENUMERATION_EXTENSION_NAME, instance_extensions);
#endif

    instance_create_info.enabledExtensionCount = static_cast<uint32_t>(instance_extensions.size());
    instance_create_info.ppEnabledExtensionNames = instance_extensions.data();

    // Enable validation layers if available
    std::vector<const char*> enabled_layer_names;
    addValidationLayerNameToVectorIfSupported("VK_LAYER_KHRONOS_validation", enabled_layer_names);
    instance_create_info.enabledLayerCount = static_cast<uint32_t>(enabled_layer_names.size());
    instance_create_info.ppEnabledLayerNames = enabled_layer_names.data();
#ifdef __APPLE__
    instance_create_info.flags = VK_INSTANCE_CREATE_ENUMERATE_PORTABILITY_BIT_KHR;
#endif

    // Create the Vulkan instance
    result = vkCreateInstance(&instance_create_info, nullptr, &vk_instance);
    VKL_CHECK_VULKAN_RESULT(result);
    if (!vk_instance) {
        VKL_EXIT_WITH_ERROR("Failed to create Vulkan instance.");
    }

    // Create a Vulkan window surface using GLFW
    result = glfwCreateWindowSurface(vk_instance, window, nullptr, &vk_surface);
    VKL_CHECK_VULKAN_RESULT(result);
    if (!vk_surface) {
        VKL_EXIT_WITH_ERROR("Failed to create Vulkan window surface.");
    }

    // Select a physical device (GPU)
    uint32_t physical_devices_count;
    vkEnumeratePhysicalDevices(vk_instance, &physical_devices_count, nullptr);
    if (physical_devices_count == 0) {
        VKL_EXIT_WITH_ERROR("No Vulkan-compatible physical devices found.");
    }

    std::vector<VkPhysicalDevice> physical_devices(physical_devices_count);
    vkEnumeratePhysicalDevices(vk_instance, &physical_devices_count, physical_devices.data());

    uint32_t selected_physical_device_index = selectPhysicalDeviceIndex(physical_devices, vk_surface);
    vk_physical_device = physical_devices[selected_physical_device_index];
    if (!vk_physical_device) {
        VKL_EXIT_WITH_ERROR("Failed to select a physical device.");
    }

    // Select a queue family that supports graphics and presentation
    std::array<float, 1> queue_priorities = { 1.0f };
    uint32_t selected_queue_family_index = selectQueueFamilyIndex(vk_physical_device, vk_surface);

    VkDeviceQueueCreateInfo device_queue_create_info = {};
    device_queue_create_info.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
    device_queue_create_info.queueFamilyIndex = selected_queue_family_index;
    device_queue_create_info.queueCount = 1u;
    device_queue_create_info.pQueuePriorities = queue_priorities.data();

    // Sanity check for valid queue family index
    uint32_t queue_family_count = 0;
    vkGetPhysicalDeviceQueueFamilyProperties(vk_physical_device, &queue_family_count, nullptr);
    if (selected_queue_family_index >= queue_family_count) {
        VKL_EXIT_WITH_ERROR("Invalid queue family index selected.");
    }

    // Create a logical device and retrieve the graphics queue
    std::vector<const char*> device_extensions;
    addDeviceExtensionToVectorIfSupported(VK_KHR_SWAPCHAIN_EXTENSION_NAME, vk_physical_device, device_extensions);
    addDeviceExtensionToVectorIfSupported(VK_KHR_SYNCHRONIZATION_2_EXTENSION_NAME, vk_physical_device, device_extensions);
    addDeviceExtensionToVectorIfSupported(VK_KHR_CREATE_RENDERPASS_2_EXTENSION_NAME, vk_physical_device, device_extensions);

    // Check for Synchronization2 support
    if (std::find(device_extensions.begin(), device_extensions.end(), VK_KHR_SYNCHRONIZATION_2_EXTENSION_NAME) != device_extensions.end()) {
        g_synchronization2_supported = true;
    }
#ifdef __APPLE__
    addDeviceExtensionToVectorIfSupported("VK_KHR_portability_subset", vk_physical_device, device_extensions);
#endif

    VkDeviceCreateInfo device_create_info = {};
    device_create_info.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
    device_create_info.queueCreateInfoCount = 1u;
    device_create_info.pQueueCreateInfos = &device_queue_create_info;
    device_create_info.enabledExtensionCount = static_cast<uint32_t>(device_extensions.size());
    device_create_info.ppEnabledExtensionNames = device_extensions.data();

    VkPhysicalDeviceFeatures enabled_physical_device_features = {};
    enabled_physical_device_features.fillModeNonSolid = VK_TRUE;
    enabled_physical_device_features.tessellationShader = VK_TRUE;
    device_create_info.pEnabledFeatures = &enabled_physical_device_features;

    // Enable Synchronization2 features if supported
    VkPhysicalDeviceSynchronization2FeaturesKHR physical_device_sync2_features = {};
    physical_device_sync2_features.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SYNCHRONIZATION_2_FEATURES_KHR;
    physical_device_sync2_features.synchronization2 = VK_TRUE;
    if (g_synchronization2_supported) {
        device_create_info.pNext = &physical_device_sync2_features;
    }

    result = vkCreateDevice(vk_physical_device, &device_create_info, nullptr, &vk_device);
    VKL_CHECK_VULKAN_RESULT(result);

    if (g_synchronization2_supported) {
        auto* procAddr = vkGetDeviceProcAddr(vk_device, "vkCmdPipelineBarrier2KHR");
        if (procAddr == nullptr) {
            g_synchronization2_supported = false;
    }
        else {
            g_vkCmdPipelineBarrier2KHR = reinterpret_cast<PFN_vkCmdPipelineBarrier2KHR>(procAddr);
        }
}
    if (!vk_device) {
        VKL_EXIT_WITH_ERROR("Failed to create logical device.");
    }

    // Get the graphics queue handle
    vkGetDeviceQueue(vk_device, selected_queue_family_index, 0u, &vk_queue);
    if (!vk_queue) {
        VKL_EXIT_WITH_ERROR("Failed to retrieve the graphics queue.");
    }

    // Create a swapchain
    uint32_t queueFamilyIndexCount = 0u;
    std::vector<uint32_t> queueFamilyIndices;
    VkSurfaceFormatKHR surface_format = getSurfaceImageFormat(vk_physical_device, vk_surface);
    queueFamilyIndices.push_back(selected_queue_family_index);
    queueFamilyIndexCount = 1u;
    VkSurfaceCapabilitiesKHR surface_capabilities = getPhysicalDeviceSurfaceCapabilities(vk_physical_device, vk_surface);

    // Build the swapchain configuration
    VkSwapchainCreateInfoKHR swapchain_create_info = {};
    swapchain_create_info.sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
    swapchain_create_info.surface = vk_surface;
    swapchain_create_info.minImageCount = surface_capabilities.minImageCount;
    swapchain_create_info.imageArrayLayers = 1u;
    swapchain_create_info.imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
    if (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_TRANSFER_SRC_BIT) {
        swapchain_create_info.imageUsage |= VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
    }
    else {
        std::cout << "Warning: VK_IMAGE_USAGE_TRANSFER_SRC_BIT image usage is not supported" << std::endl;
    }
    swapchain_create_info.preTransform = surface_capabilities.currentTransform;
    swapchain_create_info.compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
    swapchain_create_info.clipped = VK_TRUE;
    swapchain_create_info.imageSharingMode = VK_SHARING_MODE_EXCLUSIVE;
    swapchain_create_info.queueFamilyIndexCount = queueFamilyIndexCount;
    swapchain_create_info.pQueueFamilyIndices = queueFamilyIndices.data();
    swapchain_create_info.imageFormat = surface_format.format;
    swapchain_create_info.imageColorSpace = surface_format.colorSpace;
    swapchain_create_info.imageExtent = VkExtent2D{ static_cast<uint32_t>(window_width), static_cast<uint32_t>(window_height) };
    swapchain_create_info.presentMode = VK_PRESENT_MODE_FIFO_KHR;

    // Create the swapchain
    result = vkCreateSwapchainKHR(vk_device, &swapchain_create_info, nullptr, &vk_swapchain);
    VKL_CHECK_VULKAN_RESULT(result);
    if (!vk_swapchain) {
        VKL_EXIT_WITH_ERROR("Failed to create swapchain.");
    }

    // Retrieve swapchain images
    uint32_t swapchain_image_count;
    vkGetSwapchainImagesKHR(vk_device, vk_swapchain, &swapchain_image_count, nullptr);
    std::vector<VkImage> swapchain_image_handles(swapchain_image_count);
    vkGetSwapchainImagesKHR(vk_device, vk_swapchain, &swapchain_image_count, swapchain_image_handles.data());

    // Create a depth buffer
    VkImage depth_buffer = vklCreateDeviceLocalImageWithBackingMemory(
        vk_physical_device, vk_device, window_width, window_height, VK_FORMAT_D32_SFLOAT, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT);

    VkClearValue depth_clear_value;
    depth_clear_value.depthStencil.depth = 1.0f;
    depth_clear_value.depthStencil.stencil = 0u;

    // Initialize the GCG framework
    VklSwapchainConfig swapchain_config = {};

    VkClearValue color_clear_value;
    color_clear_value.color.float32[0] = BACKGROUND_R;
    color_clear_value.color.float32[1] = BACKGROUND_G;
    color_clear_value.color.float32[2] = BACKGROUND_B;
    color_clear_value.color.float32[3] = 1.0f;

    swapchain_config.swapchainHandle = vk_swapchain;
    swapchain_config.imageExtent = swapchain_create_info.imageExtent;
    for (const VkImage& img : swapchain_image_handles) {
        VklSwapchainFramebufferComposition framebufferComposition;
        framebufferComposition.colorAttachmentImageDetails.imageHandle = img;
        framebufferComposition.colorAttachmentImageDetails.imageFormat = swapchain_create_info.imageFormat;
        framebufferComposition.colorAttachmentImageDetails.imageUsage = swapchain_create_info.imageUsage;
        framebufferComposition.colorAttachmentImageDetails.clearValue = color_clear_value;
        if (depthtest) {
            framebufferComposition.depthAttachmentImageDetails.imageHandle = depth_buffer;
            framebufferComposition.depthAttachmentImageDetails.imageFormat = VK_FORMAT_D32_SFLOAT;
            framebufferComposition.depthAttachmentImageDetails.imageUsage = VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT;
            framebufferComposition.depthAttachmentImageDetails.clearValue = depth_clear_value;
        }
        swapchain_config.swapchainImages.push_back(framebufferComposition);
    }

    if (!gcgInitFramework(vk_instance, vk_surface, vk_physical_device, vk_device, vk_queue, swapchain_config)) {
        VKL_EXIT_WITH_ERROR("Failed to initialize the GCG framework.");
    }

    // Create a custom graphics pipeline
    std::vector<VkDescriptorSetLayoutBinding> descriptor_set_layout_bindings = {
        {0u, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1u,
         VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT | VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, nullptr},
        {1u, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1u,
         VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT | VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, nullptr},
        {2u, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1u,
         VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT | VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, nullptr},
        {3u, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1u,
         VK_SHADER_STAGE_FRAGMENT_BIT, nullptr},
        {4u, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1u,
         VK_SHADER_STAGE_FRAGMENT_BIT, nullptr},
        {5u, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1u,
        VK_SHADER_STAGE_COMPUTE_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, nullptr}
    };

    std::vector<VkDescriptorSetLayoutBinding> particle_descriptor_bindings = {
    {0u, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1u, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, nullptr},
    {5u, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1u, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_COMPUTE_BIT, nullptr}
    };

    // Descriptor set layout binding for the particle storage buffer
    //VkDescriptorSetLayoutBinding particleSsboBinding{};
    //particleSsboBinding.binding = 5; 
    //particleSsboBinding.descriptorType = VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
    //particleSsboBinding.descriptorCount = 1;
    //particleSsboBinding.stageFlags = VK_SHADER_STAGE_COMPUTE_BIT
    //    | VK_SHADER_STAGE_VERTEX_BIT
    //    | VK_SHADER_STAGE_FRAGMENT_BIT;
    //particleSsboBinding.pImmutableSamplers = nullptr;

    //VkDescriptorSetLayoutCreateInfo descLayoutCI{};
    //descLayoutCI.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
    //descLayoutCI.bindingCount = 1;
    //descLayoutCI.pBindings = &particleSsboBinding;

    //VkDescriptorSetLayout computeDescriptorSetLayout = VK_NULL_HANDLE;
    //vkCreateDescriptorSetLayout(vk_device, &descLayoutCI, nullptr, &computeDescriptorSetLayout);


    const size_t POLYMODES = 2;
    const size_t CULLMODES = 3;
    const size_t ILLUMODES = 1;
    VkPipeline custom_pipelines[POLYMODES][CULLMODES][ILLUMODES];
    VkPolygonMode polygon_modes[POLYMODES] = { VK_POLYGON_MODE_FILL, VK_POLYGON_MODE_LINE };
    VkCullModeFlags cull_modes[CULLMODES] = { VK_CULL_MODE_NONE, VK_CULL_MODE_BACK_BIT, VK_CULL_MODE_FRONT_BIT };
    const char* shaders[ILLUMODES][2] = { {"assets/shaders/texture.vert", "assets/shaders/texture.frag"} };

    // Create pipelines for different configurations
    for (size_t i = 0; i < POLYMODES; ++i) {
        for (size_t j = 0; j < CULLMODES; ++j) {
            std::vector<std::vector<std::string>> allShaderPaths = gcgFindAllShaderFiles<ILLUMODES, 2>(shaders);
            for (size_t k = 0; k < ILLUMODES; ++k) {
                custom_pipelines[i][j][k] = vklCreateGraphicsPipeline(
                    VklGraphicsPipelineConfig{
                        allShaderPaths[k][0].c_str(),
                        allShaderPaths[k][1].c_str(),
                        {
                            {0u, sizeof(float) * 3, VK_VERTEX_INPUT_RATE_VERTEX},
                            {1u, sizeof(float) * 3, VK_VERTEX_INPUT_RATE_VERTEX},
                            {2u, sizeof(float) * 2, VK_VERTEX_INPUT_RATE_VERTEX},
                        },
                        {
                            {0u, 0u, VK_FORMAT_R32G32B32_SFLOAT, 0u},
                            {1u, 1u, VK_FORMAT_R32G32B32_SFLOAT, 0u},
                            {2u, 2u, VK_FORMAT_R32G32_SFLOAT, 0u}
                        },
                        polygon_modes[i],
                        cull_modes[j],
                        descriptor_set_layout_bindings
                    }
                );
            }
        }
    }





    CustomRenderPassInfo renderPassInfo = createCustomRenderPassAndFramebuffers(vk_device, swapchain_config);

    VkShaderModule vertModule = createShaderModule(vk_device, "assets/shaders/texture.vert", VK_SHADER_STAGE_VERTEX_BIT);
    VkShaderModule waterVertModule = createShaderModule(vk_device, "assets/shaders/waterTexture.vert", VK_SHADER_STAGE_VERTEX_BIT);
    VkShaderModule waterTescModule = createShaderModule(vk_device, "assets/shaders/water.tesc", VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT);
    VkShaderModule waterTeseModule = createShaderModule(vk_device, "assets/shaders/water.tese", VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT);
    VkShaderModule fragModule = createShaderModule(vk_device, "assets/shaders/texture.frag", VK_SHADER_STAGE_FRAGMENT_BIT);
    VkShaderModule waterFragModule = createShaderModule(vk_device, "assets/shaders/waterTexture.frag", VK_SHADER_STAGE_FRAGMENT_BIT);
    VkShaderModule particleCompModule = createShaderModule(vk_device, "assets/shaders/particle.comp", VK_SHADER_STAGE_COMPUTE_BIT);
    VkShaderModule vertParticleModule = createShaderModule(vk_device, "assets/shaders/textureParticle.vert", VK_SHADER_STAGE_VERTEX_BIT);
    VkShaderModule fragParticleModule = createShaderModule(vk_device, "assets/shaders/textureParticle.frag", VK_SHADER_STAGE_FRAGMENT_BIT);

    // Vertex Shader
    VkPipelineShaderStageCreateInfo vertStageInfo = {};
    vertStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    vertStageInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
    vertStageInfo.module = vertModule;
    vertStageInfo.pName = "main";

    VkPipelineShaderStageCreateInfo waterVertStageInfo = {};
    waterVertStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    waterVertStageInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
    waterVertStageInfo.module = waterVertModule;
    waterVertStageInfo.pName = "main";

    // Tessellation Control
    VkPipelineShaderStageCreateInfo tescStageInfo = {};
    tescStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    tescStageInfo.stage = VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT;
    tescStageInfo.module = waterTescModule;
    tescStageInfo.pName = "main";

    // Tessellation Evaluation
    VkPipelineShaderStageCreateInfo teseStageInfo = {};
    teseStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    teseStageInfo.stage = VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT;
    teseStageInfo.module = waterTeseModule;
    teseStageInfo.pName = "main";

    // Fragment Shader
    VkPipelineShaderStageCreateInfo fragStageInfo = {};
    fragStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    fragStageInfo.stage = VK_SHADER_STAGE_FRAGMENT_BIT;
    fragStageInfo.module = fragModule;
    fragStageInfo.pName = "main";

    VkPipelineShaderStageCreateInfo waterFragStageInfo = {};
    waterFragStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    waterFragStageInfo.stage = VK_SHADER_STAGE_FRAGMENT_BIT;
    waterFragStageInfo.module = waterFragModule;
    waterFragStageInfo.pName = "main";

    // Shaders for the particle pipelines
    VkPipelineShaderStageCreateInfo particleCompStageInfo = {};
    particleCompStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    particleCompStageInfo.stage = VK_SHADER_STAGE_COMPUTE_BIT;
    particleCompStageInfo.module = particleCompModule;
    particleCompStageInfo.pName = "main";

    VkPipelineShaderStageCreateInfo vertParticleStageInfo = {};
    vertParticleStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    vertParticleStageInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
    vertParticleStageInfo.module = vertParticleModule;
    vertParticleStageInfo.pName = "main";

    VkPipelineShaderStageCreateInfo fragParticleStageInfo = {};
    fragParticleStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    fragParticleStageInfo.stage = VK_SHADER_STAGE_FRAGMENT_BIT;
    fragParticleStageInfo.module = fragParticleModule;
    fragParticleStageInfo.pName = "main";


    std::vector<VkPipelineShaderStageCreateInfo> shaderStages = {
        vertStageInfo,
        fragStageInfo
    };

    std::vector<VkPipelineShaderStageCreateInfo> shaderStagesWater = {
        waterVertStageInfo,
        tescStageInfo,
        teseStageInfo,
        waterFragStageInfo
    };

    std::vector<VkPipelineShaderStageCreateInfo> shaderStagesParticles = {
        vertParticleStageInfo,
        fragParticleStageInfo,
    };

    VkPipelineInputAssemblyStateCreateInfo inputAssembly = {};
    inputAssembly.sType = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
    inputAssembly.topology = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
    inputAssembly.primitiveRestartEnable = VK_FALSE;

    VkPipelineInputAssemblyStateCreateInfo inputAssemblyWater = {};
    inputAssemblyWater.sType = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
    inputAssemblyWater.topology = VK_PRIMITIVE_TOPOLOGY_PATCH_LIST;
    inputAssemblyWater.primitiveRestartEnable = VK_FALSE;

    VkPipelineInputAssemblyStateCreateInfo inputAssemblyParticles = {};
    inputAssemblyParticles.sType = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
    inputAssemblyParticles.topology = VK_PRIMITIVE_TOPOLOGY_POINT_LIST;
    inputAssemblyParticles.primitiveRestartEnable = VK_FALSE;

    VkPipelineTessellationStateCreateInfo tessellationState = {};
    tessellationState.sType = VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_STATE_CREATE_INFO;
    tessellationState.patchControlPoints = 3;

    // Vertex input buffers
    std::vector<VkVertexInputBindingDescription> vertexInputBuffers = {
        { 0u, sizeof(float) * 3, VK_VERTEX_INPUT_RATE_VERTEX }, 
        { 1u, sizeof(float) * 3, VK_VERTEX_INPUT_RATE_VERTEX }, 
        { 2u, sizeof(float) * 2, VK_VERTEX_INPUT_RATE_VERTEX }, 
    };

    // Vertex attributes
    std::vector<VkVertexInputAttributeDescription> inputAttributeDescriptions = {
        { 0u, 0u, VK_FORMAT_R32G32B32_SFLOAT, 0u },  
        { 1u, 1u, VK_FORMAT_R32G32B32_SFLOAT, 0u }, 
        { 2u, 2u, VK_FORMAT_R32G32_SFLOAT,     0u }, 
    };

    VkPolygonMode polygonMode = VK_POLYGON_MODE_FILL;
    VkCullModeFlags cullMode = VK_CULL_MODE_BACK_BIT;
    auto descriptorLayoutBindings = descriptor_set_layout_bindings;

    VkDescriptorSetLayoutCreateInfo dslCI{};
    dslCI.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
    dslCI.bindingCount = static_cast<uint32_t>(descriptorLayoutBindings.size());
    dslCI.pBindings = descriptorLayoutBindings.data();

    VkDescriptorSetLayout descriptorSetLayout = VK_NULL_HANDLE;
    vkCreateDescriptorSetLayout(vk_device, &dslCI, nullptr, &descriptorSetLayout);

    VkDescriptorSetLayoutCreateInfo particle_dsl_ci{};
    particle_dsl_ci.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
    particle_dsl_ci.bindingCount = static_cast<uint32_t>(particle_descriptor_bindings.size());
    particle_dsl_ci.pBindings = particle_descriptor_bindings.data();

    VkDescriptorSetLayout particle_descriptor_set_layout = VK_NULL_HANDLE;
    vkCreateDescriptorSetLayout(vk_device, &particle_dsl_ci, nullptr, &particle_descriptor_set_layout);


    VkPipelineLayoutCreateInfo pipelineLayoutCI{};
    pipelineLayoutCI.sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
    pipelineLayoutCI.setLayoutCount = 1;
    pipelineLayoutCI.pSetLayouts = &descriptorSetLayout;

    VkPipelineLayout pipelineLayout = VK_NULL_HANDLE;
    vkCreatePipelineLayout(vk_device, &pipelineLayoutCI, nullptr, &pipelineLayout);

    VkPipelineLayoutCreateInfo particle_pipeline_layout_ci{};
    particle_pipeline_layout_ci.sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
    particle_pipeline_layout_ci.setLayoutCount = 1;
    particle_pipeline_layout_ci.pSetLayouts = &particle_descriptor_set_layout;

    VkPipelineLayout particle_pipeline_layout;
    vkCreatePipelineLayout(vk_device, &particle_pipeline_layout_ci, nullptr, &particle_pipeline_layout);

    VkPipelineVertexInputStateCreateInfo vertexInput{};
    vertexInput.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
    vertexInput.vertexBindingDescriptionCount = static_cast<uint32_t>(vertexInputBuffers.size());
    vertexInput.pVertexBindingDescriptions = vertexInputBuffers.data();
    vertexInput.vertexAttributeDescriptionCount = static_cast<uint32_t>(inputAttributeDescriptions.size());
    vertexInput.pVertexAttributeDescriptions = inputAttributeDescriptions.data();

    VkPipelineVertexInputStateCreateInfo vertexInputParticles = {};
    vertexInputParticles.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
    vertexInputParticles.vertexBindingDescriptionCount = 0;  // No vertex buffers
    vertexInputParticles.pVertexBindingDescriptions = nullptr;
    vertexInputParticles.vertexAttributeDescriptionCount = 0;
    vertexInputParticles.pVertexAttributeDescriptions = nullptr;

    VkViewport viewport{};
    viewport.x = 0.0f;
    viewport.y = 0.0f;
    viewport.width = static_cast<float>(window_width);
    viewport.height = static_cast<float>(window_height);
    viewport.minDepth = 0.0f;
    viewport.maxDepth = 1.0f;

    VkRect2D scissor{};
    scissor.offset = { 0,0 };
    scissor.extent = { static_cast<uint32_t>(window_width), static_cast<uint32_t>(window_height) };

    VkPipelineViewportStateCreateInfo viewportState{};
    viewportState.sType = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
    viewportState.viewportCount = 1;
    viewportState.pViewports = &viewport;
    viewportState.scissorCount = 1;
    viewportState.pScissors = &scissor;

    VkPipelineRasterizationStateCreateInfo rasterState{};
    rasterState.sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
    rasterState.depthClampEnable = VK_FALSE;
    rasterState.rasterizerDiscardEnable = VK_FALSE;
    rasterState.polygonMode = polygonMode; 
    rasterState.cullMode = cullMode;   
    rasterState.frontFace = VK_FRONT_FACE_COUNTER_CLOCKWISE;
    rasterState.lineWidth = 1.0f;

    VkPipelineRasterizationStateCreateInfo rasterStateWater{};
    rasterStateWater.sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
    rasterStateWater.depthClampEnable = VK_FALSE;
    rasterStateWater.rasterizerDiscardEnable = VK_FALSE;
    rasterStateWater.polygonMode = polygonMode;
    rasterStateWater.cullMode = cullMode;
    rasterStateWater.frontFace = VK_FRONT_FACE_CLOCKWISE;
    rasterStateWater.lineWidth = 1.0f;

    VkPipelineRasterizationStateCreateInfo rasterStateParticles = {};
    rasterStateParticles.sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
    rasterStateParticles.depthClampEnable = VK_FALSE;
    rasterStateParticles.rasterizerDiscardEnable = VK_FALSE;
    rasterStateParticles.polygonMode = VK_POLYGON_MODE_POINT;
    rasterStateParticles.cullMode = VK_CULL_MODE_NONE;
    rasterStateParticles.frontFace = VK_FRONT_FACE_COUNTER_CLOCKWISE;
    rasterStateParticles.depthBiasEnable = VK_FALSE;
    rasterStateParticles.lineWidth = 1.0f;

    VkPipelineMultisampleStateCreateInfo multisample{};
    multisample.sType = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
    multisample.rasterizationSamples = VK_SAMPLE_COUNT_1_BIT;

    VkPipelineDepthStencilStateCreateInfo depthStencil{};
    depthStencil.sType = VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
    depthStencil.depthTestEnable = VK_TRUE;
    depthStencil.depthWriteEnable = VK_TRUE;
    depthStencil.depthCompareOp = VK_COMPARE_OP_LESS;
    depthStencil.stencilTestEnable = VK_FALSE;

    VkPipelineColorBlendAttachmentState colorBlendAttachment{};
    colorBlendAttachment.blendEnable = VK_TRUE;
    colorBlendAttachment.srcColorBlendFactor = VK_BLEND_FACTOR_SRC_ALPHA;
    colorBlendAttachment.dstColorBlendFactor = VK_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
    colorBlendAttachment.colorBlendOp = VK_BLEND_OP_ADD;
    colorBlendAttachment.srcAlphaBlendFactor = VK_BLEND_FACTOR_ONE;
    colorBlendAttachment.dstAlphaBlendFactor = VK_BLEND_FACTOR_ZERO;
    colorBlendAttachment.alphaBlendOp = VK_BLEND_OP_ADD;
    colorBlendAttachment.colorWriteMask = VK_COLOR_COMPONENT_R_BIT
        | VK_COLOR_COMPONENT_G_BIT
        | VK_COLOR_COMPONENT_B_BIT
        | VK_COLOR_COMPONENT_A_BIT;

    VkPipelineColorBlendStateCreateInfo colorBlend{};
    colorBlend.sType = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
    colorBlend.logicOpEnable = VK_FALSE;
    colorBlend.attachmentCount = 1;
    colorBlend.pAttachments = &colorBlendAttachment;


    VkGraphicsPipelineCreateInfo pipelineInfo = {};
    pipelineInfo.sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;

    pipelineInfo.stageCount = static_cast<uint32_t>(shaderStages.size());
    pipelineInfo.pStages = shaderStages.data();

    // Input + Tess
    pipelineInfo.pVertexInputState = &vertexInput;
    pipelineInfo.pInputAssemblyState = &inputAssembly;
    pipelineInfo.pTessellationState = &tessellationState;

    // Raster + viewport + etc
    pipelineInfo.pViewportState = &viewportState;
    pipelineInfo.pRasterizationState = &rasterState;
    pipelineInfo.pMultisampleState = &multisample;
    pipelineInfo.pDepthStencilState = &depthStencil;
    pipelineInfo.pColorBlendState = &colorBlend;

    // Layout + render pass
    pipelineInfo.layout = pipelineLayout;
    pipelineInfo.renderPass = renderPassInfo.renderPass;
    pipelineInfo.subpass = 0;

    // create geometry pipeline
    VkPipeline geometry_pipeline = VK_NULL_HANDLE;
    vkCreateGraphicsPipelines(vk_device, VK_NULL_HANDLE, 1, &pipelineInfo, nullptr, &geometry_pipeline);
    g_PipelineLayoutMap[geometry_pipeline] = pipelineLayout;

    // create water pipeline
    pipelineInfo.pInputAssemblyState = &inputAssemblyWater;
    pipelineInfo.stageCount = static_cast<uint32_t>(shaderStagesWater.size());
    pipelineInfo.pStages = shaderStagesWater.data();
    pipelineInfo.pRasterizationState = &rasterStateWater;

    VkPipeline water_pipeline = VK_NULL_HANDLE;
    vkCreateGraphicsPipelines(vk_device, VK_NULL_HANDLE, 1, &pipelineInfo, nullptr, &water_pipeline);
    g_PipelineLayoutMap[water_pipeline] = pipelineLayout;

    // create particle render pipeline
    pipelineInfo.stageCount = static_cast<uint32_t>(shaderStagesParticles.size());
    pipelineInfo.pStages = shaderStagesParticles.data();
    pipelineInfo.pRasterizationState = &rasterStateParticles;
    pipelineInfo.pInputAssemblyState = &inputAssemblyParticles;
    pipelineInfo.pVertexInputState = &vertexInputParticles;
    pipelineInfo.pRasterizationState = &rasterStateParticles;
    pipelineInfo.layout = particle_pipeline_layout;

    VkPipeline particle_render_pipeline = VK_NULL_HANDLE;
    vkCreateGraphicsPipelines(vk_device, VK_NULL_HANDLE, 1, &pipelineInfo, nullptr, &particle_render_pipeline);
    g_PipelineLayoutMap[particle_render_pipeline] = particle_pipeline_layout;

    VkComputePipelineCreateInfo computePipelineCI{};
    computePipelineCI.sType = VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO;
    computePipelineCI.stage = particleCompStageInfo;
    computePipelineCI.layout = particle_pipeline_layout;

    VkPipeline particleComputePipeline = VK_NULL_HANDLE;
    vkCreateComputePipelines(vk_device, VK_NULL_HANDLE, 1,
        &computePipelineCI, nullptr, &particleComputePipeline);

    // Allocate and write descriptors
    std::vector<VkDescriptorPoolSize> pool_sizes{
        {VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 24u},
        {VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 24u}
    };

    std::vector<VkDescriptorPoolSize> particle_pool_sizes = {
    {VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1},
    {VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1}
    };


    VkDescriptorPoolCreateInfo descriptor_pool_create_info = {};
    descriptor_pool_create_info.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    descriptor_pool_create_info.maxSets = 8u;
    descriptor_pool_create_info.poolSizeCount = static_cast<uint32_t>(pool_sizes.size());
    descriptor_pool_create_info.pPoolSizes = pool_sizes.data();

    VkDescriptorPoolCreateInfo particle_pool_ci = {};
    particle_pool_ci.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    particle_pool_ci.maxSets = 2u;
    particle_pool_ci.poolSizeCount = static_cast<uint32_t>(particle_pool_sizes.size());
    particle_pool_ci.pPoolSizes = particle_pool_sizes.data();


    VkDescriptorPool vk_descriptor_pool;
    result = vkCreateDescriptorPool(vk_device, &descriptor_pool_create_info, nullptr, &vk_descriptor_pool);
    VKL_CHECK_VULKAN_RESULT(result);

    VkDescriptorPool particle_descriptor_pool;
    result = vkCreateDescriptorPool(vk_device, &particle_pool_ci, nullptr, &particle_descriptor_pool);
    VKL_CHECK_VULKAN_RESULT(result);

    VkDescriptorSetLayoutCreateInfo descriptor_set_layout_create_info = {};
    descriptor_set_layout_create_info.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
    descriptor_set_layout_create_info.bindingCount = static_cast<uint32_t>(descriptor_set_layout_bindings.size());
    descriptor_set_layout_create_info.pBindings = descriptor_set_layout_bindings.data();

    VkDescriptorSetLayout vk_descriptor_set_layout;
    result = vkCreateDescriptorSetLayout(vk_device, &descriptor_set_layout_create_info, nullptr, &vk_descriptor_set_layout);
    VKL_CHECK_VULKAN_RESULT(result);

    // Create buffer for the particles
    VkBuffer particleBuffer = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(Particle) * NUM_PARTICLES, VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_TRANSFER_SRC_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT
    );

    // DS for the particle pipeline
    VkDescriptorSetAllocateInfo allocComputeDSInfo{};
    allocComputeDSInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    allocComputeDSInfo.descriptorPool = particle_descriptor_pool;
    allocComputeDSInfo.descriptorSetCount = 1;
    allocComputeDSInfo.pSetLayouts = &particle_descriptor_set_layout;

    VkDescriptorSet particle_descriptor_set = VK_NULL_HANDLE;
    vkAllocateDescriptorSets(vk_device, &allocComputeDSInfo, &particle_descriptor_set);


    // Create buffers for the light sources
    VkDeviceSize num_dirlights = 1;
    VkBuffer ub_dirlight = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(DirectionalLight) * num_dirlights, VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    DirectionalLight directional_light = { DIRLIGHT_COLOR, glm::normalize(DIRLIGHT_DIR) };
    vklCopyDataIntoHostCoherentBuffer(ub_dirlight, &directional_light, sizeof(DirectionalLight));

    VkDeviceSize num_pointlights = 1;
    VkBuffer ub_pointlight = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(PointLight) * num_pointlights, VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    PointLight point_light = { POINTLIGHT_COLOR, POINTLIGHT_POS, POINTLIGHT_ATTENUATION };
    vklCopyDataIntoHostCoherentBuffer(ub_pointlight, &point_light, sizeof(PointLight));

    // Load textures into images
    VkCommandPoolCreateInfo command_pool_create_info = {};
    command_pool_create_info.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
    command_pool_create_info.queueFamilyIndex = selected_queue_family_index;

    VkCommandPool command_pool;
    vkCreateCommandPool(vk_device, &command_pool_create_info, nullptr, &command_pool);

    ImageAndView wood_texture = loadImage(vk_device, vk_queue, command_pool, "assets/textures/terrain.dds");
    ImageAndView tiles_diffuse = loadImage(vk_device, vk_queue, command_pool, "assets/textures/tiles_diffuse.dds");
    ImageAndView water_normal_map = loadImage(vk_device, vk_queue, command_pool, "assets/textures/water_normal.dds");

    // Create a sampler
    VkSamplerCreateInfo sampler_create_info = {};
    sampler_create_info.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    sampler_create_info.magFilter = VK_FILTER_LINEAR;
    sampler_create_info.minFilter = VK_FILTER_LINEAR;
    sampler_create_info.addressModeU = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    sampler_create_info.addressModeV = VK_SAMPLER_ADDRESS_MODE_REPEAT;
    sampler_create_info.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    sampler_create_info.minLod = 0.0f;
    sampler_create_info.maxLod = VK_LOD_CLAMP_NONE;
    VkSampler sampler;
    result = vkCreateSampler(vk_device, &sampler_create_info, nullptr, &sampler);
    VKL_CHECK_VULKAN_RESULT(result);

    // Load geometries
    Geometry cliff_geometry = createAndUploadIntoGpuMemory(
        loadGeometry("assets/objects/cliff-rocky/cliff-face-simplified.obj")
    );
    Geometry fir_tree_geometry = createAndUploadIntoGpuMemory(
        loadGeometry("assets/objects/fir-tree/fir.obj")
    );
    Geometry hazelnut_geometry = createAndUploadIntoGpuMemory(
        loadGeometry("assets/objects/hazelnutbush/Hazelnut.obj")
    );
    Geometry high_rock_cliff_geometry = createAndUploadIntoGpuMemory(
        loadGeometry("assets/objects/high-rock-cliff/high-rock-cliff.obj")
    );
    Geometry maple_tree_geometry = createAndUploadIntoGpuMemory(
        loadGeometry("assets/objects/mapletree/MapleTree.obj")
    );
    Geometry terrain_geometry = createAndUploadIntoGpuMemory(
        loadGeometry("assets/objects/terrain.obj")
    );
    Geometry water_geometry = createAndUploadIntoGpuMemory(
        loadGeometry("assets/objects/water-surface/water.obj", false)
    );

    // Create uniform buffers
    VkBuffer ub_cliff = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    VkBuffer ub_fir_tree = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    VkBuffer ub_hazelnut = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    VkBuffer ub_high_rock_cliff = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    VkBuffer ub_maple_tree = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    VkBuffer ub_terrain = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );
    VkBuffer ub_water = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );

    VkBuffer ub_particles = vklCreateHostCoherentBufferWithBackingMemory(
        sizeof(UniformBuffer), VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT
    );

    std::vector<Particle> initialParticles(NUM_PARTICLES);

    int gridCols = 100;
    int gridRows = 1000;

    for (int i = 0; i < NUM_PARTICLES; i++) {
        int col = i % gridCols;
        int row = i / gridCols;
        // Spread x from -0.5 to 0.5 and z from -0.1 to 0.1
        float x = -0.5f + 1.0f * (static_cast<float>(col) / (gridCols - 1));
        float y = 5.0f;
        float z = -0.1f + 0.2f * (static_cast<float>(row) / (gridRows - 1));

        float r1 = randf(i);
        float r2 = randf(i + 1);

        float velocityY = 0.01f + (r1 - 0.5f) * 0.02f;
        float velocityZ = 0.02f + (r2 - 0.5f) * 0.01f;

        initialParticles[i].position = glm::vec4(x, y, z, 1.0f);
        initialParticles[i].velocity = glm::vec4(0.0f, velocityY, velocityZ, 0.0f);

        initialParticles[i].killPlaneHeight = -0.5f;
    }

    // Allocate descriptor sets
    VkDescriptorSet ds_cliff = allocDescriptorSet(vk_device, vk_descriptor_pool, vk_descriptor_set_layout);
    VkDescriptorSet ds_fir_tree = allocDescriptorSet(vk_device, vk_descriptor_pool, vk_descriptor_set_layout);
    VkDescriptorSet ds_hazelnut = allocDescriptorSet(vk_device, vk_descriptor_pool, vk_descriptor_set_layout);
    VkDescriptorSet ds_high_rock_cliff = allocDescriptorSet(vk_device, vk_descriptor_pool, vk_descriptor_set_layout);
    VkDescriptorSet ds_maple_tree = allocDescriptorSet(vk_device, vk_descriptor_pool, vk_descriptor_set_layout);
    VkDescriptorSet ds_terrain = allocDescriptorSet(vk_device, vk_descriptor_pool, vk_descriptor_set_layout);
    VkDescriptorSet ds_water = allocDescriptorSet(vk_device, vk_descriptor_pool, vk_descriptor_set_layout);

    // Write descriptor sets
    writeDescriptorSet(vk_device, ds_cliff, ub_cliff, ub_dirlight, ub_pointlight, wood_texture.view, sampler, water_normal_map.view, sampler);
    writeDescriptorSet(vk_device, ds_fir_tree, ub_fir_tree, ub_dirlight, ub_pointlight, wood_texture.view, sampler, water_normal_map.view, sampler);
    writeDescriptorSet(vk_device, ds_hazelnut, ub_hazelnut, ub_dirlight, ub_pointlight, wood_texture.view, sampler, water_normal_map.view, sampler);
    writeDescriptorSet(vk_device, ds_high_rock_cliff, ub_high_rock_cliff, ub_dirlight, ub_pointlight, wood_texture.view, sampler, water_normal_map.view, sampler);
    writeDescriptorSet(vk_device, ds_maple_tree, ub_maple_tree, ub_dirlight, ub_pointlight, wood_texture.view, sampler, water_normal_map.view, sampler);
    writeDescriptorSet(vk_device, ds_terrain, ub_terrain, ub_dirlight, ub_pointlight, wood_texture.view, sampler, water_normal_map.view, sampler);
    writeDescriptorSet(vk_device, ds_water, ub_water, ub_dirlight, ub_pointlight, wood_texture.view, sampler, water_normal_map.view, sampler);
    writeParticleDescriptorSet(vk_device, particle_descriptor_set, ub_particles, particleBuffer);

    // Create a camera
    Camera camera(gcgCreatePerspectiveProjectionMatrix(glm::radians(field_of_view), aspect_ratio, near_plane_distance, far_plane_distance));
    camera.setYaw(camera_yaw);
    camera.setPitch(camera_pitch);

    // Set up callbacks for user interaction
    glfwSetMouseButtonCallback(window, mouseButtonCallbackFromGlfw);
    glfwSetScrollCallback(window, scrollCallbackFromGlfw);
    glfwSetKeyCallback(window, [](GLFWwindow* glfw_window, int key, int scancode, int action, int mods) {
        if (action != GLFW_RELEASE)
            return;
        if (key == GLFW_KEY_ESCAPE) {
            glfwSetWindowShouldClose(glfw_window, true);
        }
        if (key == GLFW_KEY_F1) {
            g_polygon_mode_index = 1 - g_polygon_mode_index;
        }
        if (key == GLFW_KEY_F2) {
            g_culling_index = (g_culling_index + 1) % 3;
        }
        if (key == GLFW_KEY_N) {
            g_draw_normals = !g_draw_normals;
        }
        if (key == GLFW_KEY_T) {
            g_draw_texcoords = !g_draw_texcoords;
        }
        if (key == GLFW_KEY_M) {
            manualCameraMode = !manualCameraMode;
        }
        });

    double mouse_x, mouse_y;

    // Scaling factors for models
    float cliff_scale = 0.0f;
    float fir_tree_scale = 0.3f;
    float hazelnut_scale = 0.08f;
    float high_rock_cliff_scale = 0.17f;
    float maple_tree_scale = 0.04f;
    float terrain_scale = 15.0f;
    float water_scale = 2.0f;
    float particle_scale = 1.0f;

    vklCopyDataIntoHostCoherentBuffer(particleBuffer, initialParticles.data(), sizeof(Particle) * NUM_PARTICLES);

    VkSemaphore imageAvailableSemaphore;

    VkSemaphoreCreateInfo semaphoreInfo{};
    semaphoreInfo.sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO;
    VkResult res = vkCreateSemaphore(vk_device, &semaphoreInfo, nullptr, &imageAvailableSemaphore);
    if (res != VK_SUCCESS) {
        throw std::runtime_error("Failed to create imageAvailableSemaphore");
    }

    bool switchedCamera;

    vklEnablePipelineHotReloading(window, GLFW_KEY_F5);
    while (!glfwWindowShouldClose(window)) {
        // Handle user input
        glfwPollEvents();
        glfwGetCursorPos(window, &mouse_x, &mouse_y);

        if (manualCameraMode) {
            camera.update(mouse_x, mouse_y, g_zoom, g_dragging, g_strafing);
            switchedCamera = true;
        }
        else {
            if (switchedCamera) {
                camera.setYaw(2);
                camera.setPitch(2);
                switchedCamera = false;
            }

            camera.update(-g_time * 40, 35 + 100 * sin(g_time * 0.2),
                8 + 2.5 * sin(g_time * 0.3), true, g_strafing);
        }

        g_time = static_cast<float>(glfwGetTime());

        UniformBuffer ub_data;
        ub_data.time = g_time;
        ub_data.userInput[0] = g_draw_normals ? 1 : 0;
        ub_data.userInput[1] = g_draw_texcoords ? 1 : 0;
        ub_data.impactX = 0.0f;
        ub_data.impactY = 0.0f;

        // View-projection matrix and camera's position
        ub_data.viewProjMatrix = camera.getViewProjectionMatrix();
        ub_data.cameraPosition = glm::vec4{ camera.getPosition(), 1.0f };
        ub_data.color = { 1.0f, 1.0f, 0.0f, 1.0f };

        // Positions of models
        glm::vec3 cliff_position = glm::vec3(0.0f, -1.0f, 0.0f);
        glm::vec3 fir_tree_position = glm::vec3(5.0f, 0.0f, 0.0f);
        glm::vec3 hazelnut_position = glm::vec3(-5.0f, 0.0f, 0.0f);
        glm::vec3 high_rock_cliff_position = glm::vec3(0.0f, -1.0f, -0.6f);
        glm::vec3 maple_tree_position = glm::vec3(5.0f, 0.0f, 5.0f);
        glm::vec3 terrain_position = glm::vec3(-1.0f, -0.4f, 5.5f);
        glm::vec3 water_position = glm::vec3(0.0f, -0.5f, 3.0f);
        glm::vec3 particle_position = glm::vec3(0.0f, -0.0f, 0.0f);

        // Update cliff model
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), cliff_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(cliff_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA, KD, KS, ALPHA };
        ub_data.userInput[2] = 0;
        ub_data.userInput[3] = 1;
        vklCopyDataIntoHostCoherentBuffer(ub_cliff, &ub_data, sizeof(UniformBuffer));

        // Update fir-tree model
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), fir_tree_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(fir_tree_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA, KD, KS, ALPHA };
        ub_data.userInput[2] = 0;
        ub_data.userInput[3] = 1;
        vklCopyDataIntoHostCoherentBuffer(ub_fir_tree, &ub_data, sizeof(UniformBuffer));

        // Update hazelnut bush model
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), hazelnut_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(hazelnut_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA, KD, KS, ALPHA };
        ub_data.userInput[2] = 0;
        ub_data.userInput[3] = 1;
        vklCopyDataIntoHostCoherentBuffer(ub_hazelnut, &ub_data, sizeof(UniformBuffer));

        // Update high-rock-cliff model
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), high_rock_cliff_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(high_rock_cliff_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA, KD, KS, ALPHA };
        ub_data.userInput[2] = 0;
        ub_data.userInput[3] = 1;
        vklCopyDataIntoHostCoherentBuffer(ub_high_rock_cliff, &ub_data, sizeof(UniformBuffer));

        // Update maple tree model
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), maple_tree_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(maple_tree_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA, KD, KS, ALPHA };
        ub_data.userInput[2] = 0;
        ub_data.userInput[3] = 1;
        vklCopyDataIntoHostCoherentBuffer(ub_maple_tree, &ub_data, sizeof(UniformBuffer));

        // Update terrain model
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), terrain_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(terrain_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA, KD, KS, ALPHA };
        ub_data.userInput[2] = 0;
        ub_data.userInput[3] = 1;
        vklCopyDataIntoHostCoherentBuffer(ub_terrain, &ub_data, sizeof(UniformBuffer));

        // Update water model
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), water_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(water_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA_water, KD_water, KS_water, ALPHA };
        ub_data.userInput[2] = 1;
        ub_data.userInput[3] = 20;
        vklCopyDataIntoHostCoherentBuffer(ub_water, &ub_data, sizeof(UniformBuffer));

        // Update particles
        ub_data.modelMatrix = glm::translate(glm::mat4(1.0f), particle_position)
            * glm::scale(glm::mat4(1.0f), glm::vec3(particle_scale));
        ub_data.modelMatrixForNormals = glm::transpose(glm::inverse(ub_data.modelMatrix));
        ub_data.materialProperties = { KA, KD, KS, ALPHA };
        ub_data.userInput[2] = 0;
        ub_data.userInput[3] = 1;
        
        ub_data.viewProjMatrix = camera.getViewProjectionMatrix();
        vklCopyDataIntoHostCoherentBuffer(ub_particles, &ub_data, sizeof(UniformBuffer));

        uint32_t currentImageIndex;
        VkResult res_next = vkAcquireNextImageKHR(
            vk_device,
            vk_swapchain,
            UINT64_MAX,
            imageAvailableSemaphore,
            VK_NULL_HANDLE,
            &currentImageIndex
        );
        if (res_next != VK_SUCCESS && res != VK_SUBOPTIMAL_KHR) {
            throw std::runtime_error("Failed to acquire swapchain image!");
        }

        VkCommandBufferAllocateInfo allocInfo{};
        allocInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
        allocInfo.commandPool = command_pool;
        allocInfo.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
        allocInfo.commandBufferCount = 1;

        VkCommandBuffer commandBuffer;
        vkAllocateCommandBuffers(vk_device, &allocInfo, &commandBuffer);

        VkCommandBufferBeginInfo beginInfo{};
        beginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
        beginInfo.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;

        vkBeginCommandBuffer(commandBuffer, &beginInfo);

        // Bind compute pipeline and descriptor sets (SSBO with Particle data)
        vkCmdBindPipeline(commandBuffer, VK_PIPELINE_BIND_POINT_COMPUTE, particleComputePipeline);
        vkCmdBindDescriptorSets(
            commandBuffer,
            VK_PIPELINE_BIND_POINT_COMPUTE,
            particle_pipeline_layout,
            0,
            1, &particle_descriptor_set,
            0, nullptr
        );

        // Dispatch enough groups for all particles
        uint32_t groupCount = (NUM_PARTICLES + 255) / 256;
        vkCmdDispatch(commandBuffer, groupCount, 1, 1);

        // Pipeline Barrier to make SSBO writes visible to Vertex/Fragment
        VkMemoryBarrier memoryBarrier{};
        memoryBarrier.sType = VK_STRUCTURE_TYPE_MEMORY_BARRIER;
        memoryBarrier.srcAccessMask = VK_ACCESS_SHADER_WRITE_BIT; // from compute
        memoryBarrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;  // for vertex/frag

        vkCmdPipelineBarrier(
            commandBuffer,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,    // srcStageMask
            VK_PIPELINE_STAGE_VERTEX_SHADER_BIT |
            VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT,   // dstStageMask
            0,                                       
            1, &memoryBarrier,
            0, nullptr,
            0, nullptr
        );


        VkClearValue clearValues[2];
        clearValues[0].color = { {0.14f, 0.4f, 0.37f, 1.f} };
        clearValues[1].depthStencil = { 1.0f, 0 };

        VkRenderPassBeginInfo rpBegin{};
        rpBegin.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
        rpBegin.renderPass = renderPassInfo.renderPass;
        rpBegin.framebuffer = renderPassInfo.framebuffers[currentImageIndex];
        rpBegin.renderArea.offset = { 0, 0 };
        rpBegin.renderArea.extent = swapchain_config.imageExtent;
        rpBegin.clearValueCount = 2;
        rpBegin.pClearValues = clearValues;

        vkCmdBeginRenderPass(commandBuffer, &rpBegin, VK_SUBPASS_CONTENTS_INLINE);

        VkPipeline selected_pipeline = custom_pipelines[g_polygon_mode_index][g_culling_index][0];

        drawGeometryWithMaterial(geometry_pipeline, cliff_geometry, ds_cliff, 1u, commandBuffer);
        drawGeometryWithMaterial(geometry_pipeline, fir_tree_geometry, ds_fir_tree, 1u, commandBuffer);
        drawGeometryWithMaterial(geometry_pipeline, hazelnut_geometry, ds_hazelnut, 1u, commandBuffer);
        drawGeometryWithMaterial(geometry_pipeline, high_rock_cliff_geometry, ds_high_rock_cliff, 1u, commandBuffer);
        drawGeometryWithMaterial(geometry_pipeline, maple_tree_geometry, ds_maple_tree, 1u, commandBuffer);
        drawGeometryWithMaterial(geometry_pipeline, terrain_geometry, ds_terrain, 1u, commandBuffer);
        drawGeometryWithMaterial(water_pipeline, water_geometry, ds_water, 1u, commandBuffer);

        // Draw particles 
        vkCmdBindPipeline(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, particle_render_pipeline);
        vkCmdBindDescriptorSets(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS,
            particle_pipeline_layout,
            0, 1, &particle_descriptor_set,
            0, nullptr);

        vkCmdDraw(commandBuffer, NUM_PARTICLES, 1, 0, 0);

        vkCmdEndRenderPass(commandBuffer);
        vkEndCommandBuffer(commandBuffer);

        VkPipelineStageFlags waitStage = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;

        VkSubmitInfo submitInfo{};
        submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
        submitInfo.commandBufferCount = 1;
        submitInfo.pCommandBuffers = &commandBuffer;
        submitInfo.waitSemaphoreCount = 1;
        submitInfo.pWaitSemaphores = &imageAvailableSemaphore;
        submitInfo.pWaitDstStageMask = &waitStage;

        vkQueueSubmit(vk_queue, 1, &submitInfo, VK_NULL_HANDLE);
        vkQueueWaitIdle(vk_queue);

        // Present rendered image to the screen

        VkPresentInfoKHR presentInfo{};
        presentInfo.sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR;
        presentInfo.waitSemaphoreCount = 0;
        presentInfo.pWaitSemaphores = nullptr;
        presentInfo.swapchainCount = 1;
        presentInfo.pSwapchains = &vk_swapchain;
        presentInfo.pImageIndices = &currentImageIndex;

        presentInfo.pResults = nullptr;

        VkResult presRes = vkQueuePresentKHR(vk_queue, &presentInfo);
        if (presRes == VK_ERROR_OUT_OF_DATE_KHR || presRes == VK_SUBOPTIMAL_KHR) {
        }
        else if (presRes != VK_SUCCESS) {
            throw std::runtime_error("Failed to present swapchain image!");
        }

        if (cmdline_args.run_headless) {
            uint32_t idx = vklGetCurrentSwapChainImageIndex();
            std::string screenshot_filename = "screenshot";
            if (cmdline_args.set_filename) {
                screenshot_filename = cmdline_args.filename;
            }
            gcgSaveScreenshot(screenshot_filename, swapchain_image_handles[idx], window_width, window_height, surface_format.format,
                vk_device, vk_physical_device, vk_queue, selected_queue_family_index);
            break;
        }
    }

    // Wait for all GPU work to finish before cleaning up
    vkDeviceWaitIdle(vk_device);

    // Cleanup
    vklDestroyDeviceLocalImageAndItsBackingMemory(depth_buffer);
    vkDestroyDescriptorSetLayout(vk_device, vk_descriptor_set_layout, nullptr);
    vkDestroyDescriptorSetLayout(vk_device, particle_descriptor_set_layout, nullptr);
    vkDestroyDescriptorPool(vk_device, vk_descriptor_pool, nullptr);
    vkDestroyDescriptorPool(vk_device, particle_descriptor_pool, nullptr);

    // Destroy uniform buffers and geometries
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_cliff);
    destroyGeometryGpuMemory(cliff_geometry);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_fir_tree);
    destroyGeometryGpuMemory(fir_tree_geometry);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_hazelnut);
    destroyGeometryGpuMemory(hazelnut_geometry);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_high_rock_cliff);
    destroyGeometryGpuMemory(high_rock_cliff_geometry);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_maple_tree);
    destroyGeometryGpuMemory(maple_tree_geometry);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_terrain);
    destroyGeometryGpuMemory(terrain_geometry);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_water);
    destroyGeometryGpuMemory(water_geometry); 
    vklDestroyHostCoherentBufferAndItsBackingMemory(particleBuffer);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_particles);

    vkDestroySampler(vk_device, sampler, nullptr);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_pointlight);
    vklDestroyHostCoherentBufferAndItsBackingMemory(ub_dirlight);
    vkDestroyImageView(vk_device, tiles_diffuse.view, nullptr);
    vklDestroyDeviceLocalImageAndItsBackingMemory(tiles_diffuse.image);
    vkDestroyImageView(vk_device, water_normal_map.view, nullptr);
    vklDestroyDeviceLocalImageAndItsBackingMemory(water_normal_map.image);
    vkDestroyImageView(vk_device, wood_texture.view, nullptr);

    for (auto view : imageViews) {
        vkDestroyImageView(vk_device, view, nullptr);
    }

    vklDestroyDeviceLocalImageAndItsBackingMemory(wood_texture.image);
    vkDestroyCommandPool(vk_device, command_pool, nullptr);

    for (size_t i = 0; i < POLYMODES; ++i) {
        for (size_t j = 0; j < CULLMODES; ++j) {
            for (size_t k = 0; k < ILLUMODES; ++k) {
                vklDestroyGraphicsPipeline(custom_pipelines[i][j][k]);
            }
        }
    }

    vkDestroySemaphore(vk_device, imageAvailableSemaphore, nullptr);

    vkDestroyDescriptorSetLayout(vk_device, descriptorSetLayout, nullptr);
    vkDestroyPipeline(vk_device, water_pipeline, nullptr);
    vkDestroyPipeline(vk_device, geometry_pipeline, nullptr);
    vkDestroyPipeline(vk_device, particleComputePipeline, nullptr);
    vkDestroyPipeline(vk_device, particle_render_pipeline, nullptr);
    vkDestroyShaderModule(vk_device, waterVertModule, nullptr);
    vkDestroyShaderModule(vk_device, waterFragModule, nullptr);
    vkDestroyShaderModule(vk_device, fragModule, nullptr);
    vkDestroyShaderModule(vk_device, vertModule, nullptr);
    vkDestroyShaderModule(vk_device, waterTescModule, nullptr);
    vkDestroyShaderModule(vk_device, waterTeseModule, nullptr);
    vkDestroyShaderModule(vk_device, particleCompModule, nullptr);
    vkDestroyShaderModule(vk_device, vertParticleModule, nullptr);
    vkDestroyShaderModule(vk_device, fragParticleModule, nullptr);
    vkDestroyRenderPass(vk_device, renderPassInfo.renderPass, nullptr);
    vkDestroyPipelineLayout(vk_device, pipelineLayout, nullptr);
    vkDestroyPipelineLayout(vk_device, particle_pipeline_layout, nullptr);

    for (size_t k = 0; k < renderPassInfo.framebuffers.size(); ++k) {
        vkDestroyFramebuffer(vk_device, renderPassInfo.framebuffers[k], nullptr);
    }

    // Destroy framework and Vulkan objects
    gcgDestroyFramework();
    vkDestroySwapchainKHR(vk_device, vk_swapchain, nullptr);
    vkDestroyDevice(vk_device, nullptr);
    vkDestroySurfaceKHR(vk_instance, vk_surface, nullptr);
    vkDestroyInstance(vk_instance, nullptr);
    glfwDestroyWindow(window);
    glfwTerminate();

    return EXIT_SUCCESS;
}

/* --------------------------------------------- */
// Helper Function Definitions
/* --------------------------------------------- */

void writeParticleDescriptorSet(VkDevice device, VkDescriptorSet set,
    VkBuffer uniform_buffer, VkBuffer storage_buffer) {
    // Binding 0
    VkDescriptorBufferInfo uniform_info = {};
    uniform_info.buffer = uniform_buffer;
    uniform_info.range = VK_WHOLE_SIZE;

    // Binding 5
    VkDescriptorBufferInfo storage_info = {};
    storage_info.buffer = storage_buffer;
    storage_info.range = VK_WHOLE_SIZE;

    std::array<VkWriteDescriptorSet, 2> writes = { {
        {
            VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, // sType
            nullptr, // pNext
            set, // dstSet
            0, // dstBinding
            0, // dstArrayElement
            1, // descriptorCount
            VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, // descriptorType
            nullptr, // pImageInfo
            &uniform_info, // pBufferInfo
            nullptr // pTexelBufferView
        },
        {
            VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, // sType
            nullptr, // pNext
            set, // dstSet
            5, // dstBinding
            0, // dstArrayElement
            1, // descriptorCount
            VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, // descriptorType
            nullptr, // pImageInfo
            &storage_info, // pBufferInfo
            nullptr // pTexelBufferView
        }
    } };

    vkUpdateDescriptorSets(device, static_cast<uint32_t>(writes.size()), writes.data(), 0, nullptr);
}


CustomRenderPassInfo createCustomRenderPassAndFramebuffers(
    VkDevice device,
    const VklSwapchainConfig& swapchain_config
)
{
    CustomRenderPassInfo result{};

    bool hasColor = (swapchain_config.swapchainImages[0].colorAttachmentImageDetails.imageHandle != VK_NULL_HANDLE);
    bool hasDepth = (swapchain_config.swapchainImages[0].depthAttachmentImageDetails.imageHandle != VK_NULL_HANDLE);

    std::vector<VkAttachmentDescription> attachmentDescs;
    std::vector<VkAttachmentReference>    colorAttachmentRefs;
    VkAttachmentReference                 depthAttachmentRef = {};

    bool hasDepthAttachments = false;

    std::vector<VkImageUsageFlags> usageSet;
    std::vector<VkFormat>          formatSet;

    std::vector<VklSwapchainImageDetails> attachments0;
    if (hasColor) {
        attachments0.push_back(swapchain_config.swapchainImages[0].colorAttachmentImageDetails);
    }
    if (hasDepth) {
        attachments0.push_back(swapchain_config.swapchainImages[0].depthAttachmentImageDetails);
    }

    // Build the attachment descriptions for these
    for (size_t j = 0; j < attachments0.size(); ++j) {
        VkAttachmentDescription desc{};
        desc.flags = 0;
        desc.format = attachments0[j].imageFormat;
        desc.samples = VK_SAMPLE_COUNT_1_BIT;
        desc.loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
        bool isDepth = (attachments0[j].imageUsage & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT) != 0;
        desc.storeOp = isDepth ? VK_ATTACHMENT_STORE_OP_DONT_CARE : VK_ATTACHMENT_STORE_OP_STORE;
        desc.stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
        desc.stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
        desc.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
        desc.finalLayout = isDepth ? VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL
            : VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;

        attachmentDescs.push_back(desc);

        // If it's color:
        if (!isDepth) {
            VkAttachmentReference colorRef{};
            colorRef.attachment = static_cast<uint32_t>(attachmentDescs.size() - 1);
            colorRef.layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
            colorAttachmentRefs.push_back(colorRef);
        }
        else {
            hasDepthAttachments = true;
            depthAttachmentRef.attachment = static_cast<uint32_t>(attachmentDescs.size() - 1);
            depthAttachmentRef.layout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
        }
    }

    // Create a single subpass that references these attachments:
    VkSubpassDescription subpassDesc{};
    subpassDesc.flags = 0;
    subpassDesc.pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
    subpassDesc.inputAttachmentCount = 0;
    subpassDesc.pInputAttachments = nullptr;

    subpassDesc.colorAttachmentCount = static_cast<uint32_t>(colorAttachmentRefs.size());
    subpassDesc.pColorAttachments = colorAttachmentRefs.data();

    VkAttachmentReference depthRefIfNeeded;
    if (hasDepthAttachments) {
        subpassDesc.pDepthStencilAttachment = &depthAttachmentRef;
    }
    else {
        subpassDesc.pDepthStencilAttachment = nullptr;
    }

    std::array<VkSubpassDependency, 2> subpassDeps{};
    subpassDeps[0].srcSubpass = VK_SUBPASS_EXTERNAL;
    subpassDeps[0].dstSubpass = 0;
    subpassDeps[0].srcStageMask = VK_PIPELINE_STAGE_TRANSFER_BIT;
    subpassDeps[0].dstStageMask = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
    subpassDeps[0].srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
    subpassDeps[0].dstAccessMask = VK_ACCESS_SHADER_READ_BIT;

    subpassDeps[1].srcSubpass = 0;
    subpassDeps[1].dstSubpass = VK_SUBPASS_EXTERNAL;
    subpassDeps[1].srcStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
    subpassDeps[1].dstStageMask = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
    subpassDeps[1].srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
    subpassDeps[1].dstAccessMask = 0;

    VkRenderPassCreateInfo renderPassCI{};
    renderPassCI.sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
    renderPassCI.attachmentCount = static_cast<uint32_t>(attachmentDescs.size());
    renderPassCI.pAttachments = attachmentDescs.data();
    renderPassCI.subpassCount = 1;
    renderPassCI.pSubpasses = &subpassDesc;
    renderPassCI.dependencyCount = static_cast<uint32_t>(subpassDeps.size());
    renderPassCI.pDependencies = subpassDeps.data();

    VkRenderPass renderPass = VK_NULL_HANDLE;
    if (vkCreateRenderPass(device, &renderPassCI, nullptr, &renderPass) != VK_SUCCESS) {
        throw std::runtime_error("Failed to create render pass!");
    }

    result.renderPass = renderPass;

    for (size_t i = 0; i < swapchain_config.swapchainImages.size(); ++i) {
        // For each image, we gather the color/depth
        std::vector<VkImageView> attachments;
        if (hasColor) {
            VkImageView colorView = createImageViewForSwapchainImage(
                device,
                swapchain_config.swapchainImages[i].colorAttachmentImageDetails
            );
            attachments.push_back(colorView);
        }
        if (hasDepth) {
            VkImageView depthView = createImageViewForSwapchainImage(
                device,
                swapchain_config.swapchainImages[i].depthAttachmentImageDetails
            );
            attachments.push_back(depthView);
        }

        VkFramebufferCreateInfo fbCI{};
        fbCI.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
        fbCI.renderPass = renderPass;
        fbCI.attachmentCount = static_cast<uint32_t>(attachments.size());
        fbCI.pAttachments = attachments.data();
        fbCI.width = swapchain_config.imageExtent.width;
        fbCI.height = swapchain_config.imageExtent.height;
        fbCI.layers = 1;

        VkFramebuffer fb = VK_NULL_HANDLE;
        if (vkCreateFramebuffer(device, &fbCI, nullptr, &fb) != VK_SUCCESS) {
            throw std::runtime_error("Failed to create custom framebuffer!");
        }

        result.framebuffers.push_back(fb);
    }

    return result;
}

VkImageView createImageViewForSwapchainImage(
    VkDevice device,
    const VklSwapchainImageDetails& imgDetails
)
{
    // if it's color usage, aspect=COLOR, if it's depth usage, aspect=DEPTH
    VkImageAspectFlags aspect = (imgDetails.imageUsage & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT)
        ? VK_IMAGE_ASPECT_DEPTH_BIT
        : VK_IMAGE_ASPECT_COLOR_BIT;

    VkImageViewCreateInfo ivCI{};
    ivCI.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    ivCI.image = imgDetails.imageHandle;
    ivCI.viewType = VK_IMAGE_VIEW_TYPE_2D;
    ivCI.format = imgDetails.imageFormat;
    ivCI.components = {
        VK_COMPONENT_SWIZZLE_IDENTITY,
        VK_COMPONENT_SWIZZLE_IDENTITY,
        VK_COMPONENT_SWIZZLE_IDENTITY,
        VK_COMPONENT_SWIZZLE_IDENTITY
    };
    ivCI.subresourceRange.aspectMask = aspect;
    ivCI.subresourceRange.baseMipLevel = 0;
    ivCI.subresourceRange.levelCount = 1;
    ivCI.subresourceRange.baseArrayLayer = 0;
    ivCI.subresourceRange.layerCount = 1;

    VkImageView view;
    if (vkCreateImageView(device, &ivCI, nullptr, &view) != VK_SUCCESS) {
        throw std::runtime_error("Failed to create image view for swapchain image!");
    }

    imageViews.push_back(view);
    return view;
}


void addInstanceExtensionToVectorIfSupported(const char* extension_name, std::vector<const char*>& ref_vector) {
    VkResult result;

    // Query how many instance extensions there are:
    uint32_t instance_extension_count;
    result = vkEnumerateInstanceExtensionProperties(nullptr, &instance_extension_count, nullptr);
    VKL_CHECK_VULKAN_ERROR(result);
    VKL_RETURN_ON_ERROR(result);

    // Get all the instance extension names/properties there are:
    std::vector<VkExtensionProperties> available_instance_extensions(instance_extension_count);
    result = vkEnumerateInstanceExtensionProperties(nullptr, &instance_extension_count, available_instance_extensions.data());
    VKL_CHECK_VULKAN_ERROR(result);
    VKL_RETURN_ON_ERROR(result);

    for (const VkExtensionProperties& available_extension : available_instance_extensions) {
        if (strcmp(available_extension.extensionName, extension_name) == 0) {
            // Found the extension => Add it to the vector:
            ref_vector.push_back(extension_name);
            return;
        }
    }
}

void addValidationLayerNameToVectorIfSupported(const char* validation_layer_name, std::vector<const char*>& ref_vector) {
    VkResult result;

    // Query how many validation layers there are:
    uint32_t layer_count;
    result = vkEnumerateInstanceLayerProperties(&layer_count, nullptr);
    VKL_CHECK_VULKAN_ERROR(result);
    VKL_RETURN_ON_ERROR(result);

    // Get all the validation layer names/properties there are:
    std::vector<VkLayerProperties> available_layers(layer_count);
    result = vkEnumerateInstanceLayerProperties(&layer_count, available_layers.data());
    VKL_CHECK_VULKAN_ERROR(result);
    VKL_RETURN_ON_ERROR(result);

    for (const VkLayerProperties& available_layer : available_layers) {
        if (strcmp(available_layer.layerName, validation_layer_name) == 0) {
            // Found the validation layer => Add it to the vector:
            ref_vector.push_back(validation_layer_name);
            return;
        }
    }
}

void addDeviceExtensionToVectorIfSupported(const char* extension_name, VkPhysicalDevice physical_device, std::vector<const char*>& ref_vector) {
    VkResult result;

    // Query how many device extensions there are:
    uint32_t extensions_count;
    result = vkEnumerateDeviceExtensionProperties(physical_device, nullptr, &extensions_count, nullptr);
    VKL_CHECK_VULKAN_ERROR(result);
    VKL_RETURN_ON_ERROR(result);

    // Get all the device extensions:
    std::vector<VkExtensionProperties> available_extensions(extensions_count);
    result = vkEnumerateDeviceExtensionProperties(physical_device, nullptr, &extensions_count, available_extensions.data());
    VKL_CHECK_VULKAN_ERROR(result);
    VKL_RETURN_ON_ERROR(result);

    for (const VkExtensionProperties& available_extension : available_extensions) {
        if (strcmp(available_extension.extensionName, extension_name) == 0) {
            // Found the extension => Add it to the vector:
            ref_vector.push_back(extension_name);
            return;
        }
    }
}

uint32_t selectPhysicalDeviceIndex(const VkPhysicalDevice* physical_devices, uint32_t physical_device_count, VkSurfaceKHR surface) {
    // Iterate over all the physical devices and select one that satisfies all our requirements.
    // Our requirements are:
    //  - Must support a queue that must have both, graphics and presentation capabilities
    for (uint32_t physical_device_index = 0u; physical_device_index < physical_device_count; ++physical_device_index) {
        // Check if fillModeNonSolid is supported
        VkPhysicalDeviceFeatures features;
        vkGetPhysicalDeviceFeatures(physical_devices[physical_device_index], &features);
        if (VK_TRUE != features.fillModeNonSolid) {
            continue; // This physical device does not support it => look for a different one
        }

        // Get the number of different queue families:
        uint32_t queue_family_count = 0;
        vkGetPhysicalDeviceQueueFamilyProperties(physical_devices[physical_device_index], &queue_family_count, nullptr);

        // Get the queue families' data:
        std::vector<VkQueueFamilyProperties> queue_families(queue_family_count);
        vkGetPhysicalDeviceQueueFamilyProperties(physical_devices[physical_device_index], &queue_family_count, queue_families.data());

        for (uint32_t queue_family_index = 0u; queue_family_index < queue_family_count; ++queue_family_index) {
            // If this physical device supports a queue family which supports both, graphics and presentation
            //  => select this physical device
            if ((queue_families[queue_family_index].queueFlags & VK_QUEUE_GRAPHICS_BIT) != 0) {
                // This queue supports graphics! Let's see if it also supports presentation:
                VkBool32 presentation_supported;
                vkGetPhysicalDeviceSurfaceSupportKHR(physical_devices[physical_device_index], queue_family_index, surface, &presentation_supported);

                if (VK_TRUE == presentation_supported) {
                    // We've found a suitable physical device
                    return physical_device_index;
                }
            }
        }
    }
    VKL_EXIT_WITH_ERROR("Unable to find a suitable physical device that supports graphics and presentation on the same queue.");
}

uint32_t selectPhysicalDeviceIndex(const std::vector<VkPhysicalDevice>& physical_devices, VkSurfaceKHR surface) {
    return selectPhysicalDeviceIndex(physical_devices.data(), static_cast<uint32_t>(physical_devices.size()), surface);
}

uint32_t selectQueueFamilyIndex(VkPhysicalDevice physical_device, VkSurfaceKHR surface) {
    // Get the number of different queue families for the given physical device:
    uint32_t queue_family_count = 0;
    vkGetPhysicalDeviceQueueFamilyProperties(physical_device, &queue_family_count, nullptr);

    // Get the queue families' data:
    std::vector<VkQueueFamilyProperties> queue_families(queue_family_count);
    vkGetPhysicalDeviceQueueFamilyProperties(physical_device, &queue_family_count, queue_families.data());
    for (uint32_t queue_family_index = 0u; queue_family_index < queue_family_count; ++queue_family_index) {
        if ((queue_families[queue_family_index].queueFlags & VK_QUEUE_GRAPHICS_BIT) != 0) {
            // This queue supports graphics! Let's see if it also supports presentation:
            VkBool32 presentation_supported;
            vkGetPhysicalDeviceSurfaceSupportKHR(physical_device, queue_family_index, surface, &presentation_supported);

            if (VK_TRUE == presentation_supported) {
                // We've found a suitable queue family on the given physical device and for the given surface
                //  => return its INDEX:
                return queue_family_index;
            }
        }
    }
    VKL_EXIT_WITH_ERROR("Unable to find a suitable queue family that supports graphics and presentation on the same queue.");
}

VkSurfaceFormatKHR getSurfaceImageFormat(VkPhysicalDevice physical_device, VkSurfaceKHR surface) {
    VkResult result;

    uint32_t surface_format_count;
    result = vkGetPhysicalDeviceSurfaceFormatsKHR(physical_device, surface, &surface_format_count, nullptr);
    VKL_CHECK_VULKAN_ERROR(result);

    std::vector<VkSurfaceFormatKHR> surface_formats(surface_format_count);
    result = vkGetPhysicalDeviceSurfaceFormatsKHR(physical_device, surface, &surface_format_count, surface_formats.data());
    VKL_CHECK_VULKAN_ERROR(result);

    if (surface_formats.empty()) {
        VKL_EXIT_WITH_ERROR("Unable to find supported surface formats.");
    }

    // Prefer a RGB8/sRGB format; If we are unable to find such, just return any:
    for (const VkSurfaceFormatKHR& f : surface_formats) {
        if ((f.format == VK_FORMAT_B8G8R8A8_SRGB || f.format == VK_FORMAT_R8G8B8A8_SRGB) && f.colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR) {
            return f;
        }
    }

    return surface_formats[0];
}

VkSurfaceCapabilitiesKHR getPhysicalDeviceSurfaceCapabilities(VkPhysicalDevice physical_device, VkSurfaceKHR surface) {
    VkSurfaceCapabilitiesKHR surface_capabilities;
    VkResult result = vkGetPhysicalDeviceSurfaceCapabilitiesKHR(physical_device, surface, &surface_capabilities);
    VKL_CHECK_VULKAN_ERROR(result);
    return surface_capabilities;
}

VkSurfaceTransformFlagBitsKHR getSurfaceTransform(VkPhysicalDevice physical_device, VkSurfaceKHR surface) {
    return getPhysicalDeviceSurfaceCapabilities(physical_device, surface).currentTransform;
}


VkDescriptorSet allocDescriptorSet(VkDevice device, VkDescriptorPool descriptor_pool, VkDescriptorSetLayout descriptor_set_layout) {
    VkResult result;

    // Prepare allocation info and allocate:
    VkDescriptorSetAllocateInfo descriptor_set_allocate_info = {};
    descriptor_set_allocate_info.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    descriptor_set_allocate_info.descriptorPool = descriptor_pool;
    descriptor_set_allocate_info.descriptorSetCount = 1u;
    descriptor_set_allocate_info.pSetLayouts = &descriptor_set_layout;

    VkDescriptorSet descriptor_set;
    result = vkAllocateDescriptorSets(device, &descriptor_set_allocate_info, &descriptor_set);
    VKL_CHECK_VULKAN_RESULT(result);

    if (result < VK_SUCCESS) {
        VKL_EXIT_WITH_ERROR("Allocating a new descriptor set from the given pool and of the given layout failed.");
    }

    return descriptor_set;
}

void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer uniform_buffer) {
    // Prepare write info and write:
    VkDescriptorBufferInfo descriptor_buffer_info = {};
    descriptor_buffer_info.buffer = uniform_buffer;
    descriptor_buffer_info.offset = static_cast<VkDeviceSize>(0);
    descriptor_buffer_info.range = VK_WHOLE_SIZE;

    VkWriteDescriptorSet write_descriptor_set = {};
    write_descriptor_set.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
    write_descriptor_set.dstSet = descriptor_set;
    write_descriptor_set.dstBinding = 0u;
    write_descriptor_set.dstArrayElement = 0u;
    write_descriptor_set.descriptorCount = 1u;
    write_descriptor_set.descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
    write_descriptor_set.pBufferInfo = &descriptor_buffer_info;

    vkUpdateDescriptorSets(device, 1u, &write_descriptor_set, 0u, nullptr);
}

void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer object_data, VkBuffer directional_light_data, VkBuffer point_light_data) {
    // Write object data descriptor first:
    writeDescriptorSet(device, descriptor_set, object_data);

    // Then write the rest:

    // Prepare info about directional light data:
    VkDescriptorBufferInfo dirlight_buffer_info = {};
    dirlight_buffer_info.buffer = directional_light_data;
    dirlight_buffer_info.offset = static_cast<VkDeviceSize>(0);
    dirlight_buffer_info.range = VK_WHOLE_SIZE;

    // Prepare info about point light data:
    VkDescriptorBufferInfo pointlight_buffer_info = {};
    pointlight_buffer_info.buffer = point_light_data;
    pointlight_buffer_info.offset = static_cast<VkDeviceSize>(0);
    pointlight_buffer_info.range = VK_WHOLE_SIZE;

    std::vector<VkWriteDescriptorSet> writes = {
        VkWriteDescriptorSet{VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, nullptr, descriptor_set,
            /* dstBinding: */ 1u, 0u, 1u,
            /* descriptorType: */ VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, nullptr,
            /* pBufferInfo: */ &dirlight_buffer_info, nullptr},
        VkWriteDescriptorSet{VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, nullptr, descriptor_set,
            /* dstBinding: */ 2u, 0u, 1u,
            /* descriptorType: */ VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, nullptr,
            /* pBufferInfo: */ &pointlight_buffer_info, nullptr},
    };

    vkUpdateDescriptorSets(device, static_cast<uint32_t>(writes.size()), writes.data(), 0u, nullptr);
}

void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer object_data, VkBuffer directional_light_data, VkBuffer point_light_data,
    VkImageView image_view, VkSampler sampler) {
    // Write object data, and light source dat first:
    writeDescriptorSet(device, descriptor_set, object_data, directional_light_data, point_light_data);

    // Then write the rest:

    // Prepare info about the texture:
    VkDescriptorImageInfo diffuse_texture_info = {};
    diffuse_texture_info.imageView = image_view;
    diffuse_texture_info.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
    diffuse_texture_info.sampler = sampler;

    // Create descriptor writes for both diffuse texture and normal map
    std::vector<VkWriteDescriptorSet> writes = {
        // Diffuse texture at binding 3
        VkWriteDescriptorSet{
            VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, // sType
            nullptr,                                // pNext
            descriptor_set,                         // dstSet
            3u,                                      // dstBinding
            0u,                                      // dstArrayElement
            1u,                                      // descriptorCount
            VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,// descriptorType
            &diffuse_texture_info,                   // pImageInfo
            nullptr,                                 // pBufferInfo
            nullptr                                  // pTexelBufferView
        }
    };

    vkUpdateDescriptorSets(device, static_cast<uint32_t>(writes.size()), writes.data(), 0u, nullptr);
}

void writeDescriptorSet(VkDevice device, VkDescriptorSet descriptor_set, VkBuffer object_data, VkBuffer directional_light_data, VkBuffer point_light_data,
    VkImageView image_view, VkSampler sampler, VkImageView normalMapView, VkSampler normalMapSampler) {
    // Write object data, and light source dat first:
    writeDescriptorSet(device, descriptor_set, object_data, directional_light_data, point_light_data);

    // Then write the rest:

    // Prepare info about the texture:
    VkDescriptorImageInfo diffuse_texture_info = {};
    diffuse_texture_info.imageView = image_view;
    diffuse_texture_info.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
    diffuse_texture_info.sampler = sampler;

    // Prepare info about the normal map
    VkDescriptorImageInfo normal_map_info = {};
    normal_map_info.imageView = normalMapView;
    normal_map_info.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
    normal_map_info.sampler = normalMapSampler;

    // Create descriptor writes for both diffuse texture and normal map
    std::vector<VkWriteDescriptorSet> writes = {
        // Diffuse texture at binding 3
        VkWriteDescriptorSet{
            VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, // sType
            nullptr,                                // pNext
            descriptor_set,                         // dstSet
            3u,                                      // dstBinding
            0u,                                      // dstArrayElement
            1u,                                      // descriptorCount
            VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,// descriptorType
            &diffuse_texture_info,                   // pImageInfo
            nullptr,                                 // pBufferInfo
            nullptr                                  // pTexelBufferView
        },
        // Normal map at binding 4
        VkWriteDescriptorSet{
            VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET, // sType
            nullptr,                                // pNext
            descriptor_set,                         // dstSet
            4u,                                      // dstBinding
            0u,                                      // dstArrayElement
            1u,                                      // descriptorCount
            VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,// descriptorType
            &normal_map_info,                        // pImageInfo
            nullptr,                                 // pBufferInfo
            nullptr                                  // pTexelBufferView
        }
    };

    vkUpdateDescriptorSets(device, static_cast<uint32_t>(writes.size()), writes.data(), 0u, nullptr);
}

void drawGeometryWithMaterial(VkPipeline pipeline, const Geometry& geometry, VkDescriptorSet material, uint32_t num_instances, VkCommandBuffer command_buffer) {
    /* --------------------------------------------- */
    // Subtask 3.4: Command Buffer Recording
    /* --------------------------------------------- */

    // Get the current command buffer:
    // VkCommandBuffer cb = vklGetCurrentCommandBuffer();
    VkCommandBuffer cb = command_buffer;

    // Record binding the descriptor set for subsequent draw calls into the command buffer:
    // VkPipelineLayout pipeline_layout = vklGetLayoutForPipeline(pipeline);

    auto searchPl = g_PipelineLayoutMap.find(pipeline);
    if (g_PipelineLayoutMap.end() == searchPl) {
        throw std::runtime_error("Couldn't find the VkPipeline passed to vklGetLayoutForPipeline. ");
    }

    vkCmdBindDescriptorSets(cb, VK_PIPELINE_BIND_POINT_GRAPHICS, searchPl->second, 0u, 1u, &material, 0u, nullptr);

    // Record the draw call into the command buffer, which uses vertex and index buffers of the geometry:
    vklCmdBindPipeline(cb, VK_PIPELINE_BIND_POINT_GRAPHICS, pipeline);
    if (VK_NULL_HANDLE == geometry.colorsBuffer) {
        VkBuffer vertex_buffers[3] = {geometry.positionsBuffer, geometry.normalsBuffer, geometry.textureCoordinatesBuffer};
        VkDeviceSize offsets[3] = {0, 0, 0};
        vkCmdBindVertexBuffers(cb, 0u, 3u, vertex_buffers, offsets);
    } else {
        VkBuffer vertex_buffers[3] = {geometry.positionsBuffer, geometry.normalsBuffer, geometry.colorsBuffer};
        VkDeviceSize offsets[3] = {0, 0, 0};
        vkCmdBindVertexBuffers(cb, 0u, 3u, vertex_buffers, offsets);
    }

    vkCmdBindIndexBuffer(cb, geometry.indicesBuffer, 0, VK_INDEX_TYPE_UINT32);
    vkCmdDrawIndexed(cb, geometry.numberOfIndices, num_instances, 0u, 0u, 0u);
}



VkImageView createImageViewForImage(VkDevice device, VkImage image, VkImageViewType view_type, VkFormat format) {
    /* --------------------------------------------- */
    // Subtask 5.6: Create an Image View for each Image
    /* --------------------------------------------- */
    VkImageViewCreateInfo image_view_create_info = {};
    image_view_create_info.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    image_view_create_info.image = image;
    image_view_create_info.viewType = view_type;
    image_view_create_info.format = format;
    image_view_create_info.components.r = VK_COMPONENT_SWIZZLE_R;
    image_view_create_info.components.g = VK_COMPONENT_SWIZZLE_G;
    image_view_create_info.components.b = VK_COMPONENT_SWIZZLE_B;
    image_view_create_info.components.a = VK_COMPONENT_SWIZZLE_A;
    image_view_create_info.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    image_view_create_info.subresourceRange.baseMipLevel = 0u;
    image_view_create_info.subresourceRange.levelCount = VK_REMAINING_MIP_LEVELS;
    image_view_create_info.subresourceRange.baseArrayLayer = 0u;
    image_view_create_info.subresourceRange.layerCount = VK_IMAGE_VIEW_TYPE_CUBE == view_type ? 6u : 1u;

    VkImageView vk_image_view;
    VkResult result = vkCreateImageView(device, &image_view_create_info, nullptr, &vk_image_view);
    VKL_CHECK_VULKAN_RESULT(result);

    return vk_image_view;
}

ImageAndView loadImage(VkDevice device, VkQueue queue, VkCommandPool command_pool, std::string image_file_name) {
    std::vector<std::string> image_file_paths = gcgFindTextureFiles({image_file_name});

    /* --------------------------------------------- */
    // Subtask 5.5: Load DDS Textures into Images
    /* --------------------------------------------- */
    VklImageInfo image_info = vklGetDdsImageInfo(image_file_paths[0].c_str());
    const uint32_t image_layers = static_cast<uint32_t>(image_file_paths.size());
    assert(image_layers == 1u || image_layers == 6u);
    VkImageCreateFlags image_flags = image_layers == 6 ? VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT : VkImageCreateFlags{};
    VkResult result;

    // 0. Create the image:
    VkImage image = vklCreateDeviceLocalImageWithBackingMemory(image_info.extent.width, image_info.extent.height, image_info.imageFormat,
        VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT | VK_IMAGE_USAGE_SAMPLED_BIT, image_layers, image_flags);

    // 1. Create a command buffer and start recording
    VkCommandBufferAllocateInfo command_buffer_allocate_info = {};
    command_buffer_allocate_info.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
    command_buffer_allocate_info.commandPool = command_pool;
    command_buffer_allocate_info.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
    command_buffer_allocate_info.commandBufferCount = 1u;

    VkCommandBuffer command_buffer;
    result = vkAllocateCommandBuffers(device, &command_buffer_allocate_info, &command_buffer);
    VKL_CHECK_VULKAN_RESULT(result);

    VkCommandBufferBeginInfo command_buffer_begin_info = {};
    command_buffer_begin_info.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
    command_buffer_begin_info.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
    vkBeginCommandBuffer(command_buffer, &command_buffer_begin_info);

    // Data structures for Synchronization2:
    VkImageMemoryBarrier2KHR image_memory_barrier2 = {}; // <-- Used only in the g_synchronization2_supported case
    image_memory_barrier2.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2_KHR;
    image_memory_barrier2.image = image;
    image_memory_barrier2.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    image_memory_barrier2.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;

    VkDependencyInfoKHR dependency_info = {};
    dependency_info.sType = VK_STRUCTURE_TYPE_DEPENDENCY_INFO_KHR;
    dependency_info.imageMemoryBarrierCount = 1u;
    dependency_info.pImageMemoryBarriers = &image_memory_barrier2;

    // Data structures for classical synchronization:
    VkImageMemoryBarrier image_memory_barrier = {}; // <-- Used only in the !g_synchronization2_supported case
    image_memory_barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
    image_memory_barrier.image = image;
    image_memory_barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    image_memory_barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;

    /* --------------------------------------------- */
    // Subtask 5.10: Enable Mipmapping
    /* --------------------------------------------- */
    std::vector<VkBuffer> staging_buffers;
    uint32_t num_levels = static_cast<uint32_t>(std::log2(std::max(image_info.extent.width, image_info.extent.height))) + 1u;

    for (uint32_t layer = 0u; layer < image_layers; ++layer) {
        image_info = vklGetDdsImageInfo(image_file_paths[layer].c_str());
        uint32_t mipWidth = image_info.extent.width;
        uint32_t mipHeight = image_info.extent.height;
        for (uint32_t level = 0u; level < num_levels; ++level) {
            image_info = vklGetDdsImageLevelInfo(image_file_paths[layer].c_str(), level);
            // Sanity check to see we are loading an image with the correct size
            if (mipWidth != image_info.extent.width || mipHeight != image_info.extent.height)
                VKL_EXIT_WITH_ERROR("vklGetDdsImageLevelInfo for level " + std::to_string(level) +
                    " returned an image with width=" + std::to_string(image_info.extent.width) + ", height=" + std::to_string(image_info.extent.height) +
                    " instead of the expected width=" + std::to_string(mipWidth) + ", height=" + std::to_string(mipHeight) + ".");
            mipWidth = mipWidth > 1u ? mipWidth / 2u : 1u;
            mipHeight = mipHeight > 1u ? mipHeight / 2u : 1u;
            // Store in vector to keep alive until they are no longer needed, which is after the fence has been signaled:
            staging_buffers.push_back(vklLoadDdsImageLevelIntoHostCoherentBuffer(image_file_paths[layer].c_str(), level));

            if (g_synchronization2_supported) {
                // 2. Record an image layout transition into a format that is optimal for the image to get data copied into it:
                image_memory_barrier2.srcStageMask = VK_PIPELINE_STAGE_2_NONE_KHR;       // No need to wait on anything
                image_memory_barrier2.srcAccessMask = VK_ACCESS_2_NONE_KHR;              // Nothing required thanks to implicit memory guaranteed with host writes
                image_memory_barrier2.dstStageMask = VK_PIPELINE_STAGE_2_COPY_BIT_KHR;   // The subsequent command (which must wait) is a COPY command.
                image_memory_barrier2.dstAccessMask = VK_ACCESS_2_TRANSFER_READ_BIT_KHR; // Copy reads from the buffer. The layout transition must be available
                                                                                         // to that type of access.
                image_memory_barrier2.oldLayout = VK_IMAGE_LAYOUT_UNDEFINED;
                image_memory_barrier2.newLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
                image_memory_barrier2.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
                image_memory_barrier2.subresourceRange.baseMipLevel = level;
                image_memory_barrier2.subresourceRange.levelCount = 1u;
                image_memory_barrier2.subresourceRange.baseArrayLayer = layer;
                image_memory_barrier2.subresourceRange.layerCount = 1u;
                g_vkCmdPipelineBarrier2KHR(command_buffer, &dependency_info);
            } else { // => Synchronization2 is NOT supported => use oldschool style:
                // 2. Record an image layout transition into a format that is optimal for the image to get data copied into it:
                image_memory_barrier.srcAccessMask = {};
                image_memory_barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
                image_memory_barrier.oldLayout = VK_IMAGE_LAYOUT_UNDEFINED;
                image_memory_barrier.newLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
                image_memory_barrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
                image_memory_barrier.subresourceRange.baseMipLevel = level;
                image_memory_barrier.subresourceRange.levelCount = 1u;
                image_memory_barrier.subresourceRange.baseArrayLayer = layer;
                image_memory_barrier.subresourceRange.layerCount = 1u;
                vkCmdPipelineBarrier(command_buffer, 
                    VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT, /* --> */ VK_PIPELINE_STAGE_TRANSFER_BIT, {},
                    /* Memory barriers:        */ 0u, nullptr,
                    /* Buffer memory barriers: */ 0u, nullptr,
                    /* Image memory barriers:  */ 1u, &image_memory_barrier);
            }

            // 3. Record copying data from the (host-coherent) buffer into the (device-local) image:
            VkBufferImageCopy buffer_image_copy = {};
            buffer_image_copy.bufferOffset = 0;
            buffer_image_copy.bufferRowLength = 0u;
            buffer_image_copy.bufferImageHeight = 0u;
            buffer_image_copy.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            buffer_image_copy.imageSubresource.mipLevel = level;
            buffer_image_copy.imageSubresource.baseArrayLayer = layer;
            buffer_image_copy.imageSubresource.layerCount = 1u;
            buffer_image_copy.imageOffset.x = 0u;
            buffer_image_copy.imageOffset.y = 0u;
            buffer_image_copy.imageOffset.z = 0u;
            buffer_image_copy.imageExtent.width = image_info.extent.width;
            buffer_image_copy.imageExtent.height = image_info.extent.height;
            buffer_image_copy.imageExtent.depth = 1u;
            vkCmdCopyBufferToImage(command_buffer, staging_buffers.back(), image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1u, &buffer_image_copy);

            if (g_synchronization2_supported) {
                // 4. Record an image layout transition into a format that is optimal for rendering:
                // Re-use the VkImageMemoryBarrier from above, but modify a few parameters:
                image_memory_barrier2.srcStageMask = VK_PIPELINE_STAGE_2_COPY_BIT_KHR;    // It is the preceding COPY which must have completed before moving on.
                image_memory_barrier2.srcAccessMask = VK_ACCESS_2_TRANSFER_WRITE_BIT_KHR; // The writes of that COPY must have completed.
                image_memory_barrier2.dstStageMask =
                    VK_PIPELINE_STAGE_2_NONE_KHR; // No dst stage required in the barrier, because afterwards we sync with the fence.
                image_memory_barrier2.dstAccessMask = VK_ACCESS_2_NONE_KHR;
                image_memory_barrier2.oldLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
                image_memory_barrier2.newLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL; // Transform into right format for being sampled from in shaders.
                g_vkCmdPipelineBarrier2KHR(command_buffer, &dependency_info);
            } else { // => Synchronization2 is NOT supported => use oldschool style:
                // 4. Record an image layout transition into a format that is optimal for rendering:
                // Re-use the VkImageMemoryBarrier from above, but modify a few parameters:
                image_memory_barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
                image_memory_barrier.dstAccessMask = {};
                image_memory_barrier.oldLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
                image_memory_barrier.newLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;

                vkCmdPipelineBarrier(command_buffer, 
                    VK_PIPELINE_STAGE_TRANSFER_BIT, /* --> */ VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT, {},
                    /* Memory barriers:        */ 0u, nullptr,
                    /* Buffer memory barriers: */ 0u, nullptr,
                    /* Image memory barriers:  */ 1u, &image_memory_barrier);
            }
        }
    }

    // 5. End recording and submit to queue; use a fence to wait for the operation to complete:
    vkEndCommandBuffer(command_buffer);

    VkFenceCreateInfo fence_create_info = {};
    fence_create_info.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;

    VkFence fence;
    result = vkCreateFence(device, &fence_create_info, nullptr, &fence);
    VKL_CHECK_VULKAN_RESULT(result);

    VkSubmitInfo submit_info = {};
    submit_info.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
    submit_info.waitSemaphoreCount = 0u;
    submit_info.commandBufferCount = 1u;
    submit_info.pCommandBuffers = &command_buffer;
    submit_info.signalSemaphoreCount = 0u;
    result = vkQueueSubmit(queue, 1u, &submit_info, fence);
    VKL_CHECK_VULKAN_RESULT(result);

    // 6. Wait for the operation to complete, then clean up:
    result = vkWaitForFences(device, 1u, &fence, VK_TRUE, UINT64_MAX);
    VKL_CHECK_VULKAN_RESULT(result);

    vkDestroyFence(device, fence, nullptr);
    vkFreeCommandBuffers(device, command_pool, 1u, &command_buffer);
    for (auto staging_buffer : staging_buffers) {
        vklDestroyHostCoherentBufferAndItsBackingMemory(staging_buffer);
    }

    return ImageAndView{
        image, createImageViewForImage(device, image, image_layers == 6 ? VK_IMAGE_VIEW_TYPE_CUBE : VK_IMAGE_VIEW_TYPE_2D, image_info.imageFormat)
    };
}
