/*
 * Copyright 2023 TU Wien, Institute of Visual Computing & Human-Centered Technology.
 * This file is part of the GCG Lab Framework and must not be redistributed.
 *
 * Original version created by Lukas Gersthofer and Bernhard Steiner.
 * Vulkan edition created by Johannes Unterguggenberger (junt@cg.tuwien.ac.at).
 */

#include "ObjLoader.h"
#include "Utils.h"
#include "UtilityFunctions.cpp"
#include <glm/gtc/constants.hpp>

#undef min
#undef max

GeometryData loadGeometry(const std::string& filepath, bool verbose) {
	GeometryData data;

	tinyobj::attrib_t attrib;
	std::vector<tinyobj::shape_t> shapes;
	std::vector<tinyobj::material_t> materials;
	std::string warn, err;

	if (!tinyobj::LoadObj(&attrib, &shapes, &materials, &warn, &err, filepath.c_str())) {
		throw std::runtime_error(warn + err);
	}

	std::unordered_map<Vertex, uint32_t> uniqueVertices{};

	for (const auto& shape : shapes) {
		for (const auto& index : shape.mesh.indices) {
			Vertex vertex{};

			// Extract vertex positions
			if (index.vertex_index >= 0) {
				vertex.position = {
					attrib.vertices[3 * index.vertex_index + 0],
					attrib.vertices[3 * index.vertex_index + 1],
					attrib.vertices[3 * index.vertex_index + 2],
				};
			}

			// Extract normals
			if (index.normal_index >= 0) {
				vertex.normal = {
					attrib.normals[3 * index.normal_index + 0],
					attrib.normals[3 * index.normal_index + 1],
					attrib.normals[3 * index.normal_index + 2],
				};
			}

			// Extract texture coordinates
			if (index.texcoord_index >= 0) {
				vertex.texcoord = {
					attrib.texcoords[2 * index.texcoord_index + 0],
					attrib.texcoords[2 * index.texcoord_index + 1],
				};
			}

			// Deduplicate vertices
			if (uniqueVertices.count(vertex) == 0) {
				uniqueVertices[vertex] = static_cast<uint32_t>(data.positions.size());
				data.positions.push_back(vertex.position);
				data.normals.push_back(vertex.normal);
				data.textureCoordinates.push_back(vertex.texcoord);
			}

			data.indices.push_back(uniqueVertices[vertex]);
		}
	}

	if (verbose) {
		std::cout << "Mesh Structure:" << std::endl;
		std::cout << "----------------" << std::endl;

		std::cout << "Vertices (" << data.positions.size() << "):" << std::endl;
		for (size_t i = 0; i < data.positions.size(); ++i) {
			std::cout << "Vertex " << i << ": "
				<< "Position(" << data.positions[i].x << ", " << data.positions[i].y << ", " << data.positions[i].z << "), "
				<< "Normal(" << data.normals[i].x << ", " << data.normals[i].y << ", " << data.normals[i].z << "), "
				<< "TexCoord(" << data.textureCoordinates[i].x << ", " << data.textureCoordinates[i].y << ")"
				<< std::endl;
		}

		std::cout << "\nIndices (" << data.indices.size() << "):" << std::endl;
		for (size_t i = 0; i < data.indices.size(); i += 3) { // Assuming triangles
			if (i + 2 < data.indices.size()) {
				std::cout << "Triangle " << (i / 3) << ": "
					<< data.indices[i] << ", "
					<< data.indices[i + 1] << ", "
					<< data.indices[i + 2]
					<< std::endl;
			}
		}
	}

	return data;
}

Geometry createAndUploadIntoGpuMemory(const GeometryData& geometry_data) {
	if (geometry_data.positions.empty()) {
		VKL_EXIT_WITH_ERROR("An empty GeometryData::positions vector has been passed to createAndUploadIntoGpuMemory(...)");
	}
	if (geometry_data.indices.empty()) {
		VKL_EXIT_WITH_ERROR("An empty GeometryData::indices vector has been passed to createAndUploadIntoGpuMemory(...)");
	}

	Geometry result;

	// Create vertex positions buffer and copy data into it:
	size_t positions_buffer_byte_size = geometry_data.positions.size() * sizeof(geometry_data.positions[0]);
	result.positionsBuffer = vklCreateHostCoherentBufferAndUploadData(
		static_cast<const void*>(geometry_data.positions.data()),
		positions_buffer_byte_size,
		VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT
	);

	// Create vertex color buffer and copy data into it:
	result.colorsBuffer = VK_NULL_HANDLE;
	if (geometry_data.colors.size() > 0) {
		size_t colors_buffer_byte_size = geometry_data.colors.size() * sizeof(geometry_data.colors[0]);
		result.colorsBuffer = vklCreateHostCoherentBufferAndUploadData(
			geometry_data.colors.data(), colors_buffer_byte_size, VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT);
	}

	// Create vertex normals buffer and copy data into it:
	size_t normals_buffer_byte_size = geometry_data.normals.size() * sizeof(geometry_data.normals[0]);
	result.normalsBuffer = vklCreateHostCoherentBufferAndUploadData(
		geometry_data.normals.data(), normals_buffer_byte_size, VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT);
	// Create vertex texture coordinates buffer and copy data into it:
	size_t texture_coordinates_buffer_byte_size = geometry_data.textureCoordinates.size() * sizeof(geometry_data.textureCoordinates[0]);
	result.textureCoordinatesBuffer = vklCreateHostCoherentBufferAndUploadData(
		geometry_data.textureCoordinates.data(),
		static_cast<VkDeviceSize>(texture_coordinates_buffer_byte_size),
		VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT
	);
	// Create indices buffer and copy data into it:
	size_t indices_buffer_byte_size = geometry_data.indices.size() * sizeof(geometry_data.indices[0]);
	result.indicesBuffer = vklCreateHostCoherentBufferAndUploadData(
		geometry_data.indices.data(),
		static_cast<VkDeviceSize>(indices_buffer_byte_size),
		VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_INDEX_BUFFER_BIT
	);
	// Also store the number of indices:
	result.numberOfIndices = static_cast<uint32_t>(geometry_data.indices.size());

	return result;
}

void destroyGeometryGpuMemory(const Geometry& geometry) {
	vklDestroyHostCoherentBufferAndItsBackingMemory(geometry.indicesBuffer);
	vklDestroyHostCoherentBufferAndItsBackingMemory(geometry.textureCoordinatesBuffer);
	vklDestroyHostCoherentBufferAndItsBackingMemory(geometry.normalsBuffer);
	if (geometry.colorsBuffer != VK_NULL_HANDLE) {
		vklDestroyHostCoherentBufferAndItsBackingMemory(geometry.colorsBuffer);
	}
	vklDestroyHostCoherentBufferAndItsBackingMemory(geometry.positionsBuffer);
}

